"""
Django models specific to the student training assessment type.
"""

from django.db import IntegrityError, models, transaction
from django.utils import timezone

from submissions import api as sub_api

from .training import TrainingExample


class StudentTrainingWorkflow(models.Model):
    """
    Tracks a student's progress through the student training assessment step.
    """
    # The submission UUID of the student being trained
    submission_uuid = models.CharField(max_length=128, db_index=True, unique=True)

    # Information about the student and problem
    # This duplicates information associated with the submission itself,
    # but we include it here to make it easier to query workflows.
    # Since submissions are immutable, we can do this without
    # jeopardizing data integrity.
    student_id = models.CharField(max_length=40, db_index=True)
    item_id = models.CharField(max_length=128, db_index=True)
    course_id = models.CharField(max_length=255, db_index=True)

    class Meta:
        app_label = "assessment"

    @classmethod
    def create_workflow(cls, submission_uuid):
        """
        Create a student training workflow.

        Args:
            submission_uuid (str): The UUID of the submission from the student
                being trained.

        Returns:
            StudentTrainingWorkflow

        Raises:
            SubmissionError: There was an error retrieving the submission.

        """
        # Retrieve the student item info
        submission = sub_api.get_submission_and_student(submission_uuid)
        student_item = submission['student_item']

        # Create the workflow
        try:
            workflow, __ = cls.objects.get_or_create(
                submission_uuid=submission_uuid,
                student_id=student_item['student_id'],
                item_id=student_item['item_id'],
                course_id=student_item['course_id']
            )
            return workflow
        # If we get an integrity error, it means we've violated a uniqueness constraint
        # (someone has created this object after we checked if it existed)
        # We can therefore assume that the object exists and do nothing.
        except IntegrityError:
            pass

    @classmethod
    def get_workflow(cls, submission_uuid):
        """
        Get a student training workflow.

        Args:
            submission_uuid (str): The UUID of the submission from the student
                being trained.

        Returns:
            StudentTrainingWorkflow. None if no workflow is found.

        """
        try:
            return cls.objects.get(submission_uuid=submission_uuid)
        except cls.DoesNotExist:
            return None

    @property
    def num_completed(self):
        """
        Return the number of training examples that the
        student successfully assessed.

        Returns:
            int

        """
        return self.items.filter(completed_at__isnull=False).count()

    def next_training_example(self, examples):
        """
        Return the next training example for the student to assess.
        If the student is already working on an example, return that.
        Otherwise, choose an example the student hasn't seen
        from the list of available examples.

        Args:
            examples (list of TrainingExample): Training examples to choose from.

        Returns:
            TrainingExample or None

        """
        # Fetch all the items for this workflow from the database
        # Since Django's `select_related` does not follow reverse keys
        # we perform the filter ourselves.
        items = StudentTrainingWorkflowItem.objects.select_related(
            'training_example'
        ).filter(workflow=self)

        # If we're already working on an item, then return that item
        incomplete_items = [item for item in items if not item.is_complete]
        if incomplete_items:
            return incomplete_items[0].training_example

        # Otherwise, pick an item that we have not completed
        # from the list of examples.
        completed_examples = [
            item.training_example for item in items
        ]
        available_examples = [
            available for available in examples
            if available not in completed_examples
        ]

        # If there are no more items available, return None
        if not available_examples:
            return None
        # Otherwise, create a new workflow item for the example
        # and add it to the workflow
        order_num = len(items) + 1
        next_example = available_examples[0]

        try:
            with transaction.atomic():
                StudentTrainingWorkflowItem.objects.create(
                    workflow=self,
                    order_num=order_num,
                    training_example=next_example
                )
        # If we get an integrity error, it means we've violated a uniqueness constraint
        # (someone has created this object after we checked if it existed)
        # Since the object already exists, we don't need to do anything
        # Use the example passed into the function, because attempting to
        # retrieve the stored example would result in an race condition.
        except IntegrityError:
            pass
        return next_example

    @property
    def current_item(self):
        """
        Return the item the student is currently working on,
        or None.

        Returns:
            StudentTrainingWorkflowItem or None

        """
        next_incomplete = self.items.select_related(
            'training_example'
        ).filter(
            completed_at__isnull=True
        ).order_by('order_num')[:1]

        return None if not next_incomplete else next_incomplete[0]


class StudentTrainingWorkflowItem(models.Model):
    """
    A particular step in the training workflow.  At each step,
    a student must try assessing an example submission.

    If the student gives the same scores as the instructor,
    then the student proceeds to the next example;
    if there are no examples left, the student has
    successfully completed training.
    """
    workflow = models.ForeignKey(StudentTrainingWorkflow, related_name="items", on_delete=models.CASCADE)
    order_num = models.PositiveIntegerField()
    started_at = models.DateTimeField(auto_now_add=True)
    completed_at = models.DateTimeField(default=None, null=True)
    training_example = models.ForeignKey(TrainingExample, on_delete=models.CASCADE)

    class Meta:
        app_label = "assessment"
        ordering = ["workflow", "order_num"]
        unique_together = ('workflow', 'order_num')

    @property
    def is_complete(self):
        """
        Check whether the student has completed this workflow item.

        Returns:
            bool

        """
        return self.completed_at is not None

    def mark_complete(self):
        """
        Mark the item as complete.  Once an item is marked complete,
        it should stay complete!

        Returns:
            None

        """
        self.completed_at = timezone.now()
        self.save()

    def check_options(self, options_selected):
        """
        Compare the options that the student selected to
        the options set by the instructor in the training example.

        Args:
            options_selected (dict): Mapping of criterion names to option names.

        Returns:
            dict

        Example usage:
            >>> item.check_options({'vocabulary': 'good', 'grammar': 'poor'})
            {'vocabulary': 'excellent'}
            >>> item.check_options({'vocabulary': 'excellent', 'grammar': 'poor'})
            {}

        """
        staff_selected = self.training_example.options_selected_dict
        corrections = dict()

        for criterion_name, option_name in staff_selected.items():
            missing_option = criterion_name not in options_selected
            incorrect_option = options_selected[criterion_name] != option_name

            if missing_option or incorrect_option:
                corrections[criterion_name] = option_name

        return corrections



"""""
A mixin for staff grading.
"""


import logging

from xblock.core import XBlock
from submissions import team_api as team_sub_api

from openassessment.assessment.api import (
    staff as staff_api,
    teams as teams_api
)
from openassessment.assessment.errors import StaffAssessmentInternalError, StaffAssessmentRequestError
from openassessment.workflow import (
    api as workflow_api,
    team_api as team_workflow_api
)

from .data_conversion import clean_criterion_feedback, create_rubric_dict, verify_assessment_parameters
from .staff_area_mixin import require_course_staff

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class StaffAssessmentMixin:
    """
    This mixin is for all staff-assessment related endpoints.
    """

    def staff_assessment_exists(self, submission_uuid):
        """
        Returns True if there exists a staff assessment for the given uuid. False otherwise.
        """

        return staff_api.get_latest_staff_assessment(submission_uuid) is not None

    @XBlock.json_handler
    @require_course_staff("STUDENT_INFO")
    @verify_assessment_parameters
    def staff_assess(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Create a staff assessment from a staff submission.
        """
        if 'submission_uuid' not in data:
            return {
                'success': False, 'msg': self._("The submission ID of the submission being assessed was not found.")
            }
        if self.is_team_assignment():
            return self._team_assess(data)
        else:
            try:
                assessment = staff_api.create_assessment(
                    data['submission_uuid'],
                    self.get_student_item_dict()["student_id"],
                    data['options_selected'],
                    clean_criterion_feedback(self.rubric_criteria, data['criterion_feedback']),
                    data['overall_feedback'],
                    create_rubric_dict(self.prompts, self.rubric_criteria_with_labels)
                )
                assess_type = data.get('assess_type', 'regrade')
                self.publish_assessment_event("openassessmentblock.staff_assess", assessment, type=assess_type)
                workflow_api.update_from_assessments(
                    assessment["submission_uuid"],
                    None,
                    override_submitter_requirements=(assess_type == 'regrade')
                )
            except StaffAssessmentRequestError:
                logger.warning(
                    "An error occurred while submitting a staff assessment "
                    "for the submission {}".format(data['submission_uuid']),
                    exc_info=True
                )
                msg = self._("Your staff assessment could not be submitted.")
                return {'success': False, 'msg': msg}
            except StaffAssessmentInternalError:
                logger.exception(
                    "An error occurred while submitting a staff assessment "
                    "for the submission {}".format(data['submission_uuid']),
                )
                msg = self._("Your staff assessment could not be submitted.")
                return {'success': False, 'msg': msg}

            return {'success': True, 'msg': ""}

    def _team_assess(self, data):
        """
        Encapsulates the functionality around staff assessment for a team based assignment
        """
        try:
            team_submission = team_sub_api.get_team_submission_from_individual_submission(data['submission_uuid'])
            team_submission_uuid = team_submission['team_submission_uuid']

            assessment = teams_api.create_assessment(
                team_submission_uuid,
                self.get_student_item_dict()["student_id"],
                data['options_selected'],
                clean_criterion_feedback(self.rubric_criteria, data['criterion_feedback']),
                data['overall_feedback'],
                create_rubric_dict(self.prompts, self.rubric_criteria_with_labels)
            )
            assess_type = data.get('assess_type', 'regrade')
            self.publish_assessment_event("openassessmentblock.staff_assess", assessment[0], type=assess_type)
            team_workflow_api.update_from_assessments(
                team_submission_uuid,
                override_submitter_requirements=(assess_type == 'regrade')
            )

        except StaffAssessmentRequestError:
            logger.warning(
                "An error occurred while submitting a team assessment "
                "for the submission {}".format(data['submission_uuid']),
                exc_info=True
            )
            msg = self._("Your team assessment could not be submitted.")
            return {'success': False, 'msg': msg}
        except StaffAssessmentInternalError:
            logger.exception(
                "An error occurred while submitting a team assessment "
                "for the submission {}".format(data['submission_uuid']),
            )
            msg = self._("Your team assessment could not be submitted.")
            return {'success': False, 'msg': msg}

        return {'success': True, 'msg': ""}

    @XBlock.handler
    def render_staff_assessment(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Renders the Staff Assessment HTML section of the XBlock
        Generates the staff assessment HTML for the Open
        Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
        more information on rendering XBlock sections.
        Args:
            data (dict):
        """
        path, context_dict = self.staff_path_and_context()

        return self.render_assessment(path, context_dict)

    def staff_path_and_context(self):
        """
        Retrieve the correct template path and template context for the handler to render.
        """
        workflow = self.get_workflow_info()
        status = workflow.get('status')
        path = 'openassessmentblock/staff/oa_staff_grade.html'
        not_available_context = {
            'status_value': self._('Not Available'),
            'button_active': 'disabled=disabled aria-expanded=false',
            'step_classes': 'is--unavailable',
        }

        if status == 'cancelled':
            context = {
                'status_value': self._('Cancelled'),
                'icon_class': 'fa-exclamation-triangle',
                'step_classes': 'is--unavailable',
                'button_active': 'disabled=disabled aria-expanded=false',
            }
        elif status == 'done':  # Staff grade exists and all steps completed.
            context = {
                'status_value': self._('Complete'),
                'icon_class': 'fa-check',
                'step_classes': 'is--complete is--empty',
                'button_active': 'disabled=disabled aria-expanded=false',
            }
        elif status == 'waiting':
            # If we are in the 'waiting' workflow, this means that a staff grade cannot exist
            # (because if a staff grade did exist, we would be in 'done' regardless of whether other
            # peers have assessed). Therefore we show that we are waiting on staff to provide a grade.
            context = {
                'status_value': self._('Not Available'),
                'message_title': self._('Waiting for a Staff Grade'),
                'message_content': self._('Check back later to see if a course staff member has assessed '
                                          'your response. You will receive your grade after the assessment '
                                          'is complete.'),
                'step_classes': 'is--showing',
                'button_active': 'aria-expanded=true',
            }
        elif status is None:  # not started
            context = not_available_context
        else:  # status is 'self' or 'peer', indicating that the student still has work to do.
            if self.staff_assessment_exists(self.submission_uuid):
                context = {
                    'status_value': self._('Complete'),
                    'icon_class': 'fa-check',
                    'message_title': self._('You Must Complete the Steps Above to View Your Grade'),
                    'message_content': self._('Although a course staff member has assessed your response, '
                                              'you will receive your grade only after you have completed '
                                              'all the required steps of this problem.'),
                    'step_classes': 'is--initially--collapsed',
                    'button_active': 'aria-expanded=false',
                }
            else:  # Both student and staff still have work to do, just show "Not Available".
                context = not_available_context

        context['xblock_id'] = self.get_xblock_id()
        return path, context



"""
Grade step in the OpenAssessment XBlock.
"""


import copy

from django.utils.translation import ugettext as _

from lazy import lazy
from xblock.core import XBlock
from openassessment.assessment.errors import PeerAssessmentError, SelfAssessmentError

from .data_conversion import create_submission_dict


class GradeMixin:
    """Grade Mixin introduces all handlers for displaying grades

    Abstracts all functionality and handlers associated with Grades.

    GradeMixin is a Mixin for the OpenAssessmentBlock. Functions in the
    GradeMixin call into the OpenAssessmentBlock functions and will not work
    outside of OpenAssessmentBlock.

    """

    @XBlock.handler
    def render_grade(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Render the grade step.

        Args:
            data: Not used.

        Keyword Arguments:
            suffix: Not used.

        Returns:
            unicode: HTML content of the grade step.
        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api as sub_api

        # Retrieve the status of the workflow.  If no workflows have been
        # started this will be an empty dict, so status will be None.
        workflow = self.get_workflow_info()
        status = workflow.get('status')

        # Default context is empty
        context = {'xblock_id': self.get_xblock_id()}

        assessment_steps = self.assessment_steps
        # Render the grading section based on the status of the workflow
        try:
            if status == "cancelled":
                path = 'openassessmentblock/grade/oa_grade_cancelled.html'
                context['score'] = workflow['score']
            elif status == "done":
                path, context = self.render_grade_complete(workflow)
            elif status == "waiting":
                # The class "is--waiting--staff" is needed in the grade template for the javascript to
                # send focus to the correct step.
                # In the case where the user has completed all steps but is still waiting on a staff grade,
                # we want focus to go from the assessment steps to the staff grading step.
                if "staff-assessment" in assessment_steps:
                    context['is_waiting_staff'] = "is--waiting--staff"
                context['score_explanation'] = self._get_score_explanation(workflow)

                path = 'openassessmentblock/grade/oa_grade_waiting.html'
            elif status is None:
                path = 'openassessmentblock/grade/oa_grade_not_started.html'
            else:  # status is 'self' or 'peer', which implies that the workflow is incomplete
                path, context = self.render_grade_incomplete(workflow)
        except (sub_api.SubmissionError, PeerAssessmentError, SelfAssessmentError):
            return self.render_error(self._("An unexpected error occurred."))
        else:
            return self.render_assessment(path, context)

    def render_grade_complete(self, workflow):
        """
        Render the grade complete state.

        Args:
            workflow (dict): The serialized Workflow model.

        Returns:
            tuple of context (dict), template_path (string)
        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api
        from openassessment.assessment.api import self as self_api
        from openassessment.assessment.api import staff as staff_api
        from submissions import api as sub_api

        # Peer specific stuff...
        assessment_steps = self.assessment_steps
        submission_uuid = workflow['submission_uuid']

        staff_assessment = None
        self_assessment = None
        feedback = None
        peer_assessments = []
        has_submitted_feedback = False

        if "peer-assessment" in assessment_steps:
            peer_api.get_score(submission_uuid, self.workflow_requirements()["peer"])
            feedback = peer_api.get_assessment_feedback(submission_uuid)
            peer_assessments = [
                self._assessment_grade_context(peer_assessment)
                for peer_assessment in peer_api.get_assessments(submission_uuid)
            ]
            has_submitted_feedback = feedback is not None

        if "self-assessment" in assessment_steps:
            self_assessment = self._assessment_grade_context(
                self_api.get_assessment(submission_uuid)
            )

        raw_staff_assessment = staff_api.get_latest_staff_assessment(submission_uuid)
        if raw_staff_assessment:
            staff_assessment = self._assessment_grade_context(raw_staff_assessment)

        feedback_text = feedback.get('feedback', '') if feedback else ''
        student_submission = sub_api.get_submission(submission_uuid)

        # We retrieve the score from the workflow, which in turn retrieves
        # the score for our current submission UUID.
        # We look up the score by submission UUID instead of student item
        # to ensure that the score always matches the rubric.
        # It's possible for the score to be `None` even if the workflow status is "done"
        # when all the criteria in the rubric are feedback-only (no options).
        score = workflow['score']

        context = {
            'score': score,
            'score_explanation': self._get_score_explanation(workflow),
            'feedback_text': feedback_text,
            'has_submitted_feedback': has_submitted_feedback,
            'student_submission': create_submission_dict(student_submission, self.prompts),
            'peer_assessments': peer_assessments,
            'grade_details': self.grade_details(
                submission_uuid,
                peer_assessments=peer_assessments,
                self_assessment=self_assessment,
                staff_assessment=staff_assessment,
            ),
            'file_upload_type': self.file_upload_type,
            'allow_multiple_files': self.allow_multiple_files,
            'allow_latex': self.allow_latex,
            'prompts_type': self.prompts_type,
            'file_urls': self.get_download_urls_from_submission(student_submission),
            'xblock_id': self.get_xblock_id()
        }

        return ('openassessmentblock/grade/oa_grade_complete.html', context)

    def render_grade_incomplete(self, workflow):
        """
        Render the grade incomplete state.

        Args:
            workflow (dict): The serialized Workflow model.

        Returns:
            tuple of context (dict), template_path (string)
        """
        def _is_incomplete(step):
            return step in workflow["status_details"] and not workflow["status_details"][step]["complete"]

        incomplete_steps = []
        if _is_incomplete("peer"):
            incomplete_steps.append(self._("Peer Assessment"))
        if _is_incomplete("self"):
            incomplete_steps.append(self._("Self Assessment"))

        return (
            'openassessmentblock/grade/oa_grade_incomplete.html',
            {
                'incomplete_steps': incomplete_steps,
                'xblock_id': self.get_xblock_id(),
                'score_explanation': self._get_score_explanation(workflow)
            }
        )

    @XBlock.json_handler
    def submit_feedback(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Submit feedback on an assessment.

        Args:
            data (dict): Can provide keys 'feedback_text' (unicode) and
                'feedback_options' (list of unicode).

        Keyword Arguments:
            suffix (str): Unused

        Returns:
            Dict with keys 'success' (bool) and 'msg' (unicode)

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api

        feedback_text = data.get('feedback_text', '')
        feedback_options = data.get('feedback_options', list())

        try:
            peer_api.set_assessment_feedback({
                'submission_uuid': self.submission_uuid,
                'feedback_text': feedback_text,
                'options': feedback_options,
            })
        except (peer_api.PeerAssessmentInternalError, peer_api.PeerAssessmentRequestError):
            return {'success': False, 'msg': self._("Assessment feedback could not be saved.")}
        else:
            self.runtime.publish(
                self,
                "openassessmentblock.submit_feedback_on_assessments",
                {
                    'submission_uuid': self.submission_uuid,
                    'feedback_text': feedback_text,
                    'options': feedback_options,
                }
            )
            return {'success': True, 'msg': self._("Feedback saved.")}

    def grade_details(
            self, submission_uuid, peer_assessments, self_assessment, staff_assessment,
            is_staff=False
    ):
        # pylint: disable=unicode-format-string
        """
        Returns details about the grade assigned to the submission.

        Args:
            submission_uuid (str): The id of the submission being graded.
            peer_assessments (list of dict): Serialized assessment models from the peer API.
            self_assessment (dict): Serialized assessment model from the self API
            staff_assessment (dict): Serialized assessment model from the staff API
            is_staff (bool): True if the grade details are being displayed to staff, else False.
                Default value is False (meaning grade details are being shown to the learner).

        Returns:
            A dictionary with full details about the submission's grade.

        Example:
            {
                criteria: [{
                    'label': 'Test name',
                    'name': 'f78ac7d4ca1e4134b0ba4b40ca212e72',
                    'prompt': 'Test prompt',
                    'order_num': 2,
                    'options': [...]
                    'feedback': [
                        'Good job!',
                        'Excellent work!',
                    ]
                }],
                additional_feedback: [{
                }]
                ...
            }

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api
        from openassessment.assessment.api import self as self_api
        from openassessment.assessment.api import staff as staff_api

        criteria = copy.deepcopy(self.rubric_criteria_with_labels)

        def has_feedback(assessments):
            """
            Returns True if at least one assessment has feedback.

            Args:
                assessments: A list of assessments

            Returns:
                Returns True if at least one assessment has feedback.
            """
            return any(
                (
                    assessment and (
                        assessment.get('feedback', None) or has_feedback(assessment.get('individual_assessments', []))
                    )
                )
                for assessment in assessments
            )

        max_scores = peer_api.get_rubric_max_scores(submission_uuid)
        median_scores = None
        assessment_steps = self.assessment_steps
        if staff_assessment:
            median_scores = staff_api.get_assessment_scores_by_criteria(submission_uuid)
        elif "peer-assessment" in assessment_steps:
            median_scores = peer_api.get_assessment_median_scores(submission_uuid)
        elif "self-assessment" in assessment_steps:
            median_scores = self_api.get_assessment_scores_by_criteria(submission_uuid)

        for criterion in criteria:
            criterion_name = criterion['name']

            # Record assessment info for the current criterion
            criterion['assessments'] = self._graded_assessments(
                submission_uuid, criterion,
                assessment_steps,
                staff_assessment,
                peer_assessments,
                self_assessment,
                is_staff=is_staff,
            )

            # Record whether there is any feedback provided in the assessments
            criterion['has_feedback'] = has_feedback(criterion['assessments'])

            # Although we prevent course authors from modifying criteria post-release,
            # it's still possible for assessments created by course staff to
            # have criteria that differ from the current problem definition.
            # It's also possible to circumvent the post-release restriction
            # if course authors directly import a course into Studio.
            # If this happens, we simply leave the score blank so that the grade
            # section can render without error.
            criterion['median_score'] = median_scores.get(criterion_name, '')
            criterion['total_value'] = max_scores.get(criterion_name, '')

        return {
            'criteria': criteria,
            'additional_feedback': self._additional_feedback(
                staff_assessment=staff_assessment,
                peer_assessments=peer_assessments,
                self_assessment=self_assessment,
            ),
        }

    def _graded_assessments(
            self, submission_uuid, criterion, assessment_steps, staff_assessment, peer_assessments,
            self_assessment, is_staff=False
    ):
        """
        Returns an array of assessments with their associated grades.
        """
        def _get_assessment_part(title, feedback_title, part_criterion_name, assessment):
            """
            Returns the assessment part for the given criterion name.
            """
            if assessment:
                for part in assessment['parts']:
                    if part['criterion']['name'] == part_criterion_name:
                        part['title'] = title
                        part['feedback_title'] = feedback_title
                        return part
            return None

        # Fetch all the unique assessment parts
        criterion_name = criterion['name']
        staff_assessment_part = _get_assessment_part(
            _('Staff Grade'),
            _('Staff Comments'),
            criterion_name,
            staff_assessment
        )
        if "peer-assessment" in assessment_steps:
            peer_assessment_part = {
                'title': _('Peer Median Grade'),
                'criterion': criterion,
                'option': self._peer_median_option(submission_uuid, criterion),
                'individual_assessments': [
                    _get_assessment_part(
                        _('Peer {peer_index}').format(peer_index=index + 1),
                        _('Peer Comments'),
                        criterion_name,
                        peer_assessment
                    )
                    for index, peer_assessment in enumerate(peer_assessments)
                ],
            }
        else:
            peer_assessment_part = None
        self_assessment_part = _get_assessment_part(
            _('Self Assessment Grade') if is_staff else _('Your Self Assessment'),
            _('Your Comments'),  # This is only used in the LMS student-facing view
            criterion_name,
            self_assessment
        )

        # Now collect together all the assessments
        assessments = []
        if staff_assessment_part:
            assessments.append(staff_assessment_part)
        if peer_assessment_part:
            assessments.append(peer_assessment_part)
        if self_assessment_part:
            assessments.append(self_assessment_part)

        # Include points only for the first assessment
        if assessments:
            first_assessment = assessments[0]
            option = first_assessment['option']
            if option and option.get('points', None) is not None:
                first_assessment['points'] = option['points']

        return assessments

    def _peer_median_option(self, submission_uuid, criterion):
        """
        Returns the option for the median peer grade.

        Args:
            submission_uuid (str): The id for the submission.
            criterion (dict): The criterion in question.

        Returns:
            The option for the median peer grade.

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api

        median_scores = peer_api.get_assessment_median_scores(submission_uuid)
        median_score = median_scores.get(criterion['name'], None)
        median_score = -1 if not median_score else median_score

        def median_options():
            """
            Returns a list of options that should be shown to represent the median.

            Some examples:
              1. Options A=1, B=3, and C=5, a median score of 3 returns [B].
              2. Options A=1, B=3, and C=5, a median score of 4 returns [B, C].
              3. Options A=1, B=1, and C=3, a median score of 1 returns [A, B]
              4. Options A=1, B=1, C=3, and D=3, a median score of 2 return [A, B, C, D]
              5. Options A=1, B=3 and C=5, a median score of 6 returns [C]
                 Note: 5 should not happen as a median should never be out of range.
            """
            last_score = -1
            median_options = []

            # Sort the options first by name and then by points, so that if there
            # are options with identical points they will sort alphabetically rather
            # than randomly. Note that this depends upon sorted being a stable sort.
            alphabetical_options = sorted(criterion['options'], key=lambda option: option['label'])
            ordered_options = sorted(alphabetical_options, key=lambda option: option['points'])

            for option in ordered_options:
                current_score = option['points']

                # If we have reached a new score, then decide what to do next
                if current_score != last_score:

                    # If the last score we saw was already larger than the median
                    # score, then we must have collected enough so return all
                    # the median options.
                    if last_score >= median_score:
                        return median_options

                    # If the current score is exactly the median or is less,
                    # then we don't need any previously collected scores.
                    if current_score <= median_score:
                        median_options = []

                    # Update the last score to be the current one
                    last_score = current_score

                # Collect the current option in case it is applicable
                median_options.append(option)
            return median_options

        # Calculate the full list of matching options for the median, and then:
        #  - If zero or one matches are found, then just return None or the single item.
        #  - If more than one match is found, return a dict with an aggregate label,
        #  - the median score, and no explanation (it is too verbose to show an aggregate).
        options = median_options()
        if not options:
            # If we weren't able to get a median option when there should be one, show the following message
            # This happens when there are less than must_be_graded_by assessments made for the user
            if criterion['options']:
                return {'label': _('Waiting for peer reviews')}
            return None
        if len(options) == 1:
            return options[0]
        return {
            'label': ' / '.join([option['label'] for option in options]),
            'points': median_score if median_score != -1 else None,
            'explanation': None,
        }

    def _additional_feedback(self, staff_assessment, peer_assessments, self_assessment):
        """
        Returns an array of additional feedback for the specified assessments.

        Args:
            staff_assessment: The staff assessment
            peer_assessments: An array of peer assessments
            self_assessment: The self assessment

        Returns:
            Returns an array of additional feedback per assessment.
        """
        additional_feedback = []
        if staff_assessment:
            feedback = staff_assessment.get('feedback')
            if feedback:
                additional_feedback.append({
                    'title': _('Staff Comments'),
                    'feedback': feedback
                })
        if peer_assessments and len(peer_assessments) >= self.workflow_requirements()['peer']['must_be_graded_by']:
            individual_feedback = []
            for peer_index, peer_assessment in enumerate(peer_assessments):
                individual_feedback.append({
                    'title': _('Peer {peer_index}').format(peer_index=peer_index + 1),
                    'feedback': peer_assessment.get('feedback')
                })
            if any(assessment_feedback['feedback'] for assessment_feedback in individual_feedback):
                additional_feedback.append({
                    'title': _('Peer'),
                    'individual_assessments': individual_feedback
                })
        if self_assessment:
            feedback = self_assessment.get('feedback')
            if feedback:
                additional_feedback.append({
                    'title': _('Your Comments'),
                    'feedback': feedback
                })

        return additional_feedback if additional_feedback else None

    @lazy
    def _criterion_and_option_labels(self):
        """
        Retrieve criteria and option labels from the rubric in the XBlock problem definition,
        defaulting to the name value if no label is available (backwards compatibility).

        Evaluated lazily, so it will return a cached value if called repeatedly.
        For the grade mixin, this should be okay, since we can't change the problem
        definition in the LMS (the settings fields are read-only).

        Returns:
            Tuple of dictionaries:
                `criterion_labels` maps criterion names to criterion labels.
                `option_labels` maps (criterion name, option name) tuples to option labels.

        """
        criterion_labels = {}
        option_labels = {}
        for criterion in self.rubric_criteria_with_labels:
            criterion_labels[criterion['name']] = criterion['label']
            for option in criterion['options']:
                option_label_key = (criterion['name'], option['name'])
                option_labels[option_label_key] = option['label']

        return criterion_labels, option_labels

    def _assessment_grade_context(self, assessment):
        """
        Sanitize an assessment dictionary into a format that can be
        passed into the grade complete Django template.

        Args:
            assessment (dict): The serialized assessment model.

        Returns:
            dict

        """
        assessment = copy.deepcopy(assessment)

        # Retrieve dictionaries mapping criteria/option names to the associated labels.
        # This is a lazy property, so we can call it repeatedly for each assessment.
        criterion_labels, option_labels = self._criterion_and_option_labels

        # Backwards compatibility: We used to treat "name" as both a user-facing label
        # and a unique identifier for criteria and options.
        # Now we treat "name" as a unique identifier, and we've added an additional "label"
        # field that we display to the user.
        # If criteria/options in the problem definition do NOT have a "label" field
        # (because they were created before this change),
        # we create a new label that has the same value as "name".
        if assessment is not None:
            for part in assessment['parts']:
                criterion_label_key = part['criterion']['name']
                part['criterion']['label'] = criterion_labels.get(criterion_label_key, part['criterion']['name'])

                # We need to be a little bit careful here: some assessment parts
                # have only written feedback, so they're not associated with any options.
                # If that's the case, we don't need to add the label field.
                if part.get('option') is not None:
                    option_label_key = (part['criterion']['name'], part['option']['name'])
                    part['option']['label'] = option_labels.get(option_label_key, part['option']['name'])

        return assessment

    def _get_assessment_type(self, workflow):
        """
        Determine which assessment is decisive in determining the grade.
        Args:
            workflow (dict): The serialized Workflow model.
        Returns:
            str: Type of decisive assessment. Possible values are self, staff, peer.
        """
        score = workflow['score']
        complete = score is not None

        if "staff-assessment" in self.assessment_steps:
            return "staff"

        # Edge case: staff overrides the grade.
        # If a score is overriden by staff, it'll always have an
        # attached annotation type with the `staff_defined` value,
        # so we look for that in this problem's annotation and
        # return staff if it's found.
        grade_annotation_types = [annotation['annotation_type'] for annotation in (score or {}).get("annotations", [])]
        if complete and "staff_defined" in grade_annotation_types:
            return "staff"

        # For other cases, we just need to figure out the
        # priority of each (either peer or self).
        # Just loop over the values and return the first one
        # after staff.
        for _assessment_type in workflow["assessment_score_priority"]:
            # assessment_step would always have staff in it, so skip it
            # while checking the priority here.
            if _assessment_type == "staff":
                continue

            if "{}-assessment".format(_assessment_type) in self.assessment_steps:
                return _assessment_type

        return None  # Just to make pylint happy

    def _get_score_explanation(self, workflow):
        """
        Return a string which explains how grade is calculated for an ORA assessment
        (which is complete i.e all assessments have been done) based on assessment_steps.
        Args:
            workflow (dict): The serialized Workflow model.
        Returns:
            str: Message explainaing how grade is determined.
        """
        score = workflow['score']
        complete = score is not None

        assessment_type = self._get_assessment_type(workflow)

        sentences = {
            "staff": _("The grade for this problem is determined by your Staff Grade."),
            "peer": _(
                "The grade for this problem is determined by the median score of "
                "your Peer Assessments."
            ),
            "self": _("The grade for this problem is determined by your Self Assessment.")
        }
        second_sentence = sentences.get(assessment_type, "")

        if complete:
            first_sentence = _(
                "You have successfully completed this problem and received a {earned_points}/{total_points}."
            ).format(earned_points=score["points_earned"], total_points=score["points_possible"])
        else:
            first_sentence = ""
            # Special Case i.e If the submission only have peer assessment
            if "peer-assessment" in self.assessment_steps and "self-assessment" not in self.assessment_steps and \
               "staff-assessment" not in self.assessment_steps:
                first_sentence = _(
                    "You have not yet received all necessary peer reviews to determine your final grade."
                )

        return "{} {}".format(first_sentence, second_sentence).strip()

    def generate_report_data(self, user_state_iterator, limit_responses=None):
        """
        Return a list of student responses and assessments for this block in a readable way.

        Arguments:
            user_state_iterator: iterator over UserStateClient objects.
                E.g. the result of user_state_client.iter_all_for_block(block_key)
            limit_responses (int|None): maximum number of responses to include.
                Set to None (default) to include all.
        Returns:
            each call yields a tuple like:
                ("my_username", {
                    'Submission ID': 'c6551...',
                    'Item ID': 5,
                    'Anonymized Student ID': 'c801..',
                    'Assessment ID': 4,
                    'Assessment Scored Date': '2020-02-01',
                    'Assessment Scored Time': '10:03:07.218280+00:00',
                    'Assessment Type': 'PE',
                    'Anonymous Scorer Id': '6e9a...',
                    'Criterion 1: Ideas": 'Poor',
                    'Points 1': 0,
                    'Median Score 1': 0,
                    'Feedback 1': 'Does not answer the question.',
                    'Criterion 2: Content": 'Excellent',
                    'Points 2': 3,
                    'Median Score 2': 3.0,
                    'Feedback 2': 'Well described.',
                    'Criteria Count': 'Well described.',
                    'Overall Feedback': 'try again',
                    'Date/Time Final Score Given': 2020-02-01 10:03:07.218280+00:00',,
                    'Final Score Points Earned': 1,
                    'Final Score Points Possible': 5,
                    'Feedback Statements Selected': "",
                    'Feedback on Assessment': "",
                    'Response files': 'http://lms.url/...',
                    'Response': '{"file_descriptions"...}',
                    'Assessment scored At': 2020-02-01 10:03:07.218280+00:00',,
                })
        """
        from openassessment.data import OraAggregateData

        xblock_id = self.get_xblock_id()
        num_rows = 0
        for user_state in user_state_iterator:
            submission_uuid = user_state.state.get('submission_uuid')
            for row in OraAggregateData.generate_assessment_data(xblock_id, submission_uuid):
                num_rows += 1
                yield (user_state.username, row)

            if limit_responses is not None and num_rows >= limit_responses:
                # End the iterator here
                break



""" Filesystem backend for file upload. """


from pathlib import Path

from django.conf import settings
import django.core.cache
from django.urls import reverse
from django.utils.encoding import smart_text

from .. import exceptions
from .base import BaseBackend


class Backend(BaseBackend):
    """
    Upload openassessment student files to a local filesystem. Note
    that in order to use this file storage backend, you need to include the
    urls from openassessment.fileupload in your urls.py file:

    E.g:
        url(r'^openassessment/storage', include(openassessment.fileupload.urls)),

    The ORA2_FILEUPLOAD_CACHE_NAME setting will also have to be defined for the
    name of the django.core.cache instance which will maintain the list of
    active storage URLs.

    E.g:

        ORA2_FILEUPLOAD_CACHE_NAME = "ora2-storage"
        CACHES = {
            ...
            'ora2-storage': {
                'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
                ...
            },
            ...
        }
    """

    def get_upload_url(self, key, content_type):
        make_upload_url_available(self._get_key_name(key), self.UPLOAD_URL_TIMEOUT)
        return self._get_url(key)

    def get_download_url(self, key):
        key_name = self._get_key_name(key)
        if self._file_exists(key_name):
            make_download_url_available(key_name, self.DOWNLOAD_URL_TIMEOUT)
            return self._get_url(key)
        return None

    def remove_file(self, key):
        from openassessment.fileupload.views_filesystem import safe_remove, get_file_path
        return safe_remove(get_file_path(self._get_key_name(key)))

    def _get_url(self, key):
        key_name = self._get_key_name(key)
        url = reverse("openassessment-filesystem-storage", kwargs={'key': key_name})
        return url

    def _file_exists(self, key_name):
        from openassessment.fileupload.views_filesystem import get_file_path

        file_path = get_file_path(key_name)
        file_path = Path(file_path)
        return file_path.exists()


def get_cache():
    """
    Returns a django.core.cache instance in charge of maintaining the
    authorized upload and download URL.

    Raises:
        FileUploadInternalError if the cache name setting is not defined.
        InvalidCacheBackendError if the corresponding cache backend has not
        been configured.
    """
    cache_name = getattr(settings, "ORA2_FILEUPLOAD_CACHE_NAME", None)
    if cache_name is None:
        raise exceptions.FileUploadInternalError("Undefined cache backend for file upload")
    return django.core.cache.caches[cache_name]


def make_upload_url_available(url_key_name, timeout):
    """
    Authorize an upload URL.

    Arguments:
        url_key_name (str): key that uniquely identifies the upload url
        timeout (int): time in seconds before the url expires
    """
    get_cache().set(
        smart_text(get_upload_cache_key(url_key_name)),
        1, timeout
    )


def make_download_url_available(url_key_name, timeout):
    """
    Authorize a download URL.

    Arguments:
        url_key_name (str): key that uniquely identifies the url
        timeout (int): time in seconds before the url expires
    """
    get_cache().set(
        smart_text(get_download_cache_key(url_key_name)),
        1, timeout
    )


def is_upload_url_available(url_key_name):
    """
    Return True if the corresponding upload URL is available.
    """
    return get_cache().get(smart_text(get_upload_cache_key(url_key_name))) is not None


def is_download_url_available(url_key_name):
    """
    Return True if the corresponding download URL is available.
    """
    return get_cache().get(smart_text(get_download_cache_key(url_key_name))) is not None


def get_upload_cache_key(url_key_name):
    if isinstance(url_key_name, bytes):
        url_key_name = url_key_name.decode('utf-8')
    return "upload/" + url_key_name


def get_download_cache_key(url_key_name):
    if isinstance(url_key_name, bytes):
        url_key_name = url_key_name.decode('utf-8')
    return "download/" + url_key_name



"""
Django models specific to peer assessment.

NOTE: We've switched to migrations, so if you make any edits to this file, you
need to then generate a matching migration for it using:

    ./manage.py schemamigration openassessment.assessment --auto

"""


from datetime import timedelta
import logging
import random

from django.db import DatabaseError, models
from django.utils.timezone import now

from openassessment.assessment.errors import PeerAssessmentInternalError, PeerAssessmentWorkflowError
from openassessment.assessment.models.base import Assessment

logger = logging.getLogger("openassessment.assessment.models")  # pylint: disable=invalid-name


class AssessmentFeedbackOption(models.Model):
    """
    Option a student can select to provide feedback on the feedback they received.

    `AssessmentFeedback` stands in a one-to-many relationship with `AssessmentFeedbackOption`s:
    a student can select zero or more `AssessmentFeedbackOption`s when providing feedback.

    Over time, we may decide to add, delete, or reword assessment feedback options.
    To preserve data integrity, we will always get-or-create `AssessmentFeedbackOption`s
    based on the option text.
    """
    text = models.CharField(max_length=255, unique=True)

    class Meta:
        app_label = "assessment"

    def __str__(self):
        return f'"{self.text}"'


class AssessmentFeedback(models.Model):
    """
    Feedback on feedback.  When students receive their grades, they
    can provide feedback on how they were assessed, to be reviewed by course staff.

    This consists of free-form written feedback
    ("Please provide any thoughts or comments on the feedback you received from your peers")
    as well as zero or more feedback options
    ("Please select the statements below that reflect what you think of this peer grading experience")
    """
    MAXSIZE = 1024 * 100     # 100KB

    submission_uuid = models.CharField(max_length=128, unique=True, db_index=True)
    assessments = models.ManyToManyField(Assessment, related_name='assessment_feedback', default=None)
    feedback_text = models.TextField(max_length=10000, default="")
    options = models.ManyToManyField(AssessmentFeedbackOption, related_name='assessment_feedback', default=None)

    class Meta:
        app_label = "assessment"

    def add_options(self, selected_options):
        """
        Select feedback options for this assessment.
        Students can select zero or more options.

        Note: you *must* save the model before calling this method.

        Args:
            option_text_list (list of unicode): List of options that the user selected.

        Raises:
            DatabaseError
        """
        # First, retrieve options that already exist
        options = list(AssessmentFeedbackOption.objects.filter(text__in=selected_options))

        # If there are additional options that do not yet exist, create them
        new_options = [text for text in selected_options if text not in [opt.text for opt in options]]
        for new_option_text in new_options:
            options.append(AssessmentFeedbackOption.objects.create(text=new_option_text))

        # Add all options to the feedback model
        # Note that we've already saved each of the AssessmentFeedbackOption models, so they have primary keys
        # (required for adding to a many-to-many relationship)
        self.options.add(*options)  # pylint:disable=E1101


class PeerWorkflow(models.Model):
    """Internal Model for tracking Peer Assessment Workflow

    This model can be used to determine the following information required
    throughout the Peer Assessment Workflow:

    1) Get next submission that requires assessment.
    2) Does a submission have enough assessments?
    3) Has a student completed enough assessments?
    4) Does a student already have a submission open for assessment?
    5) Close open assessments when completed.
    6) Should 'over grading' be allowed for a submission?

    The student item is the author of the submission.  Peer Workflow Items are
    created for each assessment made by this student.

    """
    # Amount of time before a lease on a submission expires
    TIME_LIMIT = timedelta(hours=8)

    student_id = models.CharField(max_length=40, db_index=True)
    item_id = models.CharField(max_length=128, db_index=True)
    course_id = models.CharField(max_length=255, db_index=True)
    submission_uuid = models.CharField(max_length=128, db_index=True, unique=True)
    created_at = models.DateTimeField(default=now, db_index=True)
    completed_at = models.DateTimeField(null=True, db_index=True)
    grading_completed_at = models.DateTimeField(null=True, db_index=True)
    cancelled_at = models.DateTimeField(null=True, db_index=True)

    class Meta:
        ordering = ["created_at", "id"]
        app_label = "assessment"

    @property
    def is_cancelled(self):
        """
        Check if workflow is cancelled.

        Returns:
            True/False
        """
        return bool(self.cancelled_at)

    @classmethod
    def get_by_submission_uuid(cls, submission_uuid):
        """
        Retrieve the Peer Workflow associated with the given submission UUID.

        Args:
            submission_uuid (str): The string representation of the UUID belonging
                to the associated Peer Workflow.

        Returns:
            workflow (PeerWorkflow): The most recent peer workflow associated with
                this submission UUID.

        Raises:
            PeerAssessmentWorkflowError: Thrown when no workflow can be found for
                the associated submission UUID. This should always exist before a
                student is allow to request submissions for peer assessment.

        Examples:
            >>> PeerWorkflow.get_workflow_by_submission_uuid("abc123")
            {
                'student_id': u'Bob',
                'item_id': u'type_one',
                'course_id': u'course_1',
                'submission_uuid': u'1',
                'created_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 668850, tzinfo=<UTC>)
            }

        """
        try:
            return cls.objects.get(submission_uuid=submission_uuid)
        except cls.DoesNotExist:
            return None
        except DatabaseError as ex:
            error_message = (
                "Error finding workflow for submission UUID {}. Workflow must be "
                "created for submission before beginning peer assessment."
            ).format(submission_uuid)
            logger.exception(error_message)
            raise PeerAssessmentWorkflowError(error_message) from ex

    @classmethod
    def create_item(cls, scorer_workflow, submission_uuid):
        """
        Create a new peer workflow for a student item and submission.

        Args:
            scorer_workflow (PeerWorkflow): The peer workflow associated with the scorer.
            submission_uuid (str): The submission associated with this workflow.

        Raises:
            PeerAssessmentInternalError: Raised when there is an internal error
                creating the Workflow.
        """
        peer_workflow = cls.get_by_submission_uuid(submission_uuid)

        try:
            workflow_items = PeerWorkflowItem.objects.filter(
                scorer=scorer_workflow,
                author=peer_workflow,
                submission_uuid=submission_uuid
            )

            if workflow_items:
                item = workflow_items[0]
            else:
                item = PeerWorkflowItem.objects.create(
                    scorer=scorer_workflow,
                    author=peer_workflow,
                    submission_uuid=submission_uuid
                )
            item.started_at = now()
            item.save()
            return item
        except DatabaseError as ex:
            error_message = (
                "An internal error occurred while creating a new peer workflow "
                "item for workflow {}"
            ).format(scorer_workflow)
            logger.exception(error_message)
            raise PeerAssessmentInternalError(error_message) from ex

    def find_active_assessments(self):
        """Given a student item, return an active assessment if one is found.

        Before retrieving a new submission for a peer assessor, check to see if that
        assessor already has a submission out for assessment. If an unfinished
        assessment is found that has not expired or has not been cancelled,
        return the associated submission.

        TODO: If a user begins an assessment, then resubmits, this will never find
        the unfinished assessment. Is this OK?

        Args:
            workflow (PeerWorkflow): See if there is an associated active assessment
                for this PeerWorkflow.

        Returns:
            (PeerWorkflowItem) The PeerWorkflowItem for the submission that the
                student has open for active assessment.

        """
        oldest_acceptable = now() - self.TIME_LIMIT
        items = list(self.graded.all().select_related('author').order_by("-started_at", "-id"))
        valid_open_items = []
        completed_sub_uuids = []
        # First, remove all completed items.
        for item in items:
            if item.assessment is not None or item.author.is_cancelled:
                completed_sub_uuids.append(item.submission_uuid)
            else:
                valid_open_items.append(item)

        # Remove any open items which have a submission which has been completed.
        for item in valid_open_items:
            if (item.started_at < oldest_acceptable) or (item.submission_uuid in completed_sub_uuids):
                valid_open_items.remove(item)

        return valid_open_items[0] if valid_open_items else None

    def get_submission_for_review(self, graded_by):
        """
        Find a submission for peer assessment. This function will find the next
        submission that requires assessment, excluding any submission that has been
        completely graded, or is actively being reviewed by other students.

        Args:
            graded_by (unicode): Student ID of the scorer.

        Returns:
            submission_uuid (str): The submission_uuid for the submission to review.

        Raises:
            PeerAssessmentInternalError: Raised when there is an error retrieving
                the workflows or workflow items for this request.

        """
        timeout = (now() - self.TIME_LIMIT).strftime("%Y-%m-%d %H:%M:%S")
        # The follow query behaves as the Peer Assessment Queue. This will
        # find the next submission (via PeerWorkflow) in this course / question
        # that:
        #  1) Does not belong to you
        #  2) Does not have enough completed assessments
        #  3) Is not something you have already scored.
        #  4) Does not have a combination of completed assessments or open
        #     assessments equal to or more than the requirement.
        #  5) Has not been cancelled.
        try:
            peer_workflows = list(PeerWorkflow.objects.raw(
                "select pw.id, pw.submission_uuid "
                "from assessment_peerworkflow pw "
                "where pw.item_id=%s "
                "and pw.course_id=%s "
                "and pw.student_id<>%s "
                "and pw.grading_completed_at is NULL "
                "and pw.cancelled_at is NULL "
                "and pw.id not in ("
                "   select pwi.author_id "
                "   from assessment_peerworkflowitem pwi "
                "   where pwi.scorer_id=%s "
                "   and pwi.assessment_id is not NULL "
                ") "
                "and ("
                "   select count(pwi.id) as c "
                "   from assessment_peerworkflowitem pwi "
                "   where pwi.author_id=pw.id "
                "   and (pwi.assessment_id is not NULL or pwi.started_at > %s) "
                ") < %s "
                "order by pw.created_at, pw.id "
                "limit 1; ",
                [
                    self.item_id,
                    self.course_id,
                    self.student_id,
                    self.id,
                    timeout,
                    graded_by
                ]
            ))
            if not peer_workflows:
                return None

            return peer_workflows[0].submission_uuid
        except DatabaseError as ex:
            error_message = (
                "An internal error occurred while retrieving a peer submission "
                "for learner {}"
            ).format(self)
            logger.exception(error_message)
            raise PeerAssessmentInternalError(error_message) from ex

    def get_submission_for_over_grading(self):
        """
        Retrieve the next submission uuid for over grading in peer assessment.
        """
        # The follow query behaves as the Peer Assessment Over Grading Queue. This
        # will find a random submission (via PeerWorkflow) in this course / question
        # that:
        #  1) Does not belong to you
        #  2) Is not something you have already scored
        #  3) Has not been cancelled.
        try:
            query = list(PeerWorkflow.objects.raw(
                "select pw.id, pw.submission_uuid "
                "from assessment_peerworkflow pw "
                "where course_id=%s "
                "and item_id=%s "
                "and student_id<>%s "
                "and pw.cancelled_at is NULL "
                "and pw.id not in ( "
                "select pwi.author_id "
                "from assessment_peerworkflowitem pwi "
                "where pwi.scorer_id=%s"
                "); ",
                [self.course_id, self.item_id, self.student_id, self.id]
            ))
            workflow_count = len(query)
            if workflow_count < 1:
                return None

            random_int = random.randint(0, workflow_count - 1)
            random_workflow = query[random_int]

            return random_workflow.submission_uuid
        except DatabaseError as ex:
            error_message = (
                "An internal error occurred while retrieving a peer submission "
                "for learner {}"
            ).format(self)
            logger.exception(error_message)
            raise PeerAssessmentInternalError(error_message) from ex

    def close_active_assessment(self, submission_uuid, assessment, num_required_grades):
        """
        Updates a workflow item on the student's workflow with the associated
        assessment. When a workflow item has an assessment, it is considered
        finished.

        Args:
            submission_uuid (str): The submission the scorer is grading.
            assessment (PeerAssessment): The associate assessment for this action.
            graded_by (int): The required number of grades the peer workflow
                requires to be considered complete.

        Returns:
            None

        """
        try:
            item_query = self.graded.filter(
                submission_uuid=submission_uuid
            ).order_by("-started_at", "-id")
            items = list(item_query[:1])
            if not items:
                msg = (
                    "No open assessment was found for learner {} while assessing "
                    "submission UUID {}."
                ).format(self.student_id, submission_uuid)
                raise PeerAssessmentWorkflowError(msg)
            item = items[0]
            item.assessment = assessment
            item.save()

            if not item.author.grading_completed_at:
                if item.author.graded_by.filter(assessment__isnull=False).count() >= num_required_grades:
                    item.author.grading_completed_at = now()
                    item.author.save()

        except (DatabaseError, PeerWorkflowItem.DoesNotExist) as ex:
            error_message = (
                "An internal error occurred while retrieving a workflow item for "
                "learner {}. Workflow Items are created when submissions are "
                "pulled for assessment."
            ).format(self.student_id)
            logger.exception(error_message)
            raise PeerAssessmentWorkflowError(error_message) from ex

    def num_peers_graded(self):
        """
        Returns the number of peers the student owning the workflow has graded.

        Returns:
            integer

        """
        return self.graded.filter(assessment__isnull=False).count()

    def __repr__(self):
        return (
            "PeerWorkflow(student_id={0.student_id}, item_id={0.item_id}, "
            "course_id={0.course_id}, submission_uuid={0.submission_uuid}"
            "created_at={0.created_at}, completed_at={0.completed_at})"
        ).format(self)

    def __str__(self):
        return repr(self)


class PeerWorkflowItem(models.Model):
    """Represents an assessment associated with a particular workflow

    Created every time a submission is requested for peer assessment. The
    associated workflow represents the scorer of the given submission, and the
    assessment represents the completed assessment for this work item.

    """
    scorer = models.ForeignKey(PeerWorkflow, related_name='graded', on_delete=models.CASCADE)
    author = models.ForeignKey(PeerWorkflow, related_name='graded_by', on_delete=models.CASCADE)
    submission_uuid = models.CharField(max_length=128, db_index=True)
    started_at = models.DateTimeField(default=now, db_index=True)
    assessment = models.ForeignKey(Assessment, null=True, on_delete=models.CASCADE)

    # This WorkflowItem was used to determine the final score for the Workflow.
    scored = models.BooleanField(default=False)

    @classmethod
    def get_scored_assessments(cls, submission_uuid):
        """
        Return all scored assessments for a given submission.

        Args:
            submission_uuid (str): The UUID of the submission.

        Returns:
            QuerySet of Assessment objects.

        """
        return Assessment.objects.filter(
            pk__in=[
                item.assessment.pk for item in PeerWorkflowItem.objects.filter(
                    submission_uuid=submission_uuid, scored=True
                )
            ]
        )

    class Meta:
        ordering = ["started_at", "id"]
        app_label = "assessment"

    def __repr__(self):
        return (
            "PeerWorkflowItem(scorer={0.scorer}, author={0.author}, "
            "submission_uuid={0.submission_uuid}, "
            "started_at={0.started_at}, assessment={0.assessment}, "
            "scored={0.scored})"
        ).format(self)

    def __str__(self):
        return repr(self)



"""
Serialize and deserialize OpenAssessment XBlock content to/from XML.
"""


import json
import logging
from uuid import uuid4 as uuid

import dateutil.parser
import defusedxml.ElementTree as safe_etree
import pytz

import lxml.etree as etree
from openassessment.xblock.data_conversion import update_assessments_format
from openassessment.xblock.lms_mixin import GroupAccessDict

log = logging.getLogger(__name__)


class UpdateFromXmlError(Exception):
    """
    Error occurred while deserializing the OpenAssessment XBlock content from XML.
    """


class ValidationError(UpdateFromXmlError):
    """
    The XML definition is not semantically valid.
    """


def _sort_by_order_num(items):
    """
    Sort dictionaries by the key "order_num".
    If no order number is specified, assign an arbitrary order.
    Ignores non-dict items in the list.

    Args:
        items (list): List of dictionaries to sort.

    Returns:
        dict
    """
    return sorted(
        [el for el in items if isinstance(el, dict)],
        key=lambda el: el.get('order_num', 0)
    )


def _safe_get_text(element):
    """
    Retrieve the text from the element, safely handling empty elements.

    Args:
        element (lxml.etree.Element): The XML element.

    Returns:
        unicode
    """
    return str(element.text) if element.text is not None else ""


def _serialize_prompts(prompts_root, prompts_list):
    """
    Serialize prompts as XML, adding children to the XML with root
    node `prompts_root`.

    Args:
        prompts_root (lxml.etree.Element): The root node of the tree.
        prompts_list (list): List of prompt dictionaries.

    Returns:
        None
    """
    if not isinstance(prompts_list, list):
        return

    for prompt in prompts_list:
        prompt_el = etree.SubElement(prompts_root, 'prompt')

        # Prompt description
        prompt_description = etree.SubElement(prompt_el, 'description')
        prompt_description.text = str(prompt.get('description', ''))


def _serialize_options(options_root, options_list):
    """
    Serialize rubric criterion options as XML, adding children to the XML
    with root node `options_root`.

    We don't make any assumptions about the contents of `options_list`,
    and we handle unexpected inputs gracefully.

    Args:
        options_root (lxml.etree.Element): The root node of the tree.
        options_list (list): List of options dictionaries.

    Returns:
        None
    """
    # Sort the options by order number, then serialize as XML
    for option in _sort_by_order_num(options_list):
        option_el = etree.SubElement(options_root, 'option')

        # Points (default to 0)
        option_el.set('points', str(option.get('points', 0)))

        # Name (default to a UUID)
        option_name = etree.SubElement(option_el, 'name')
        if 'name' in option:
            option_name.text = str(option['name'])
        else:
            option_name.text = str(uuid().hex)

        # Label (default to the option name, then an empty string)
        option_label = etree.SubElement(option_el, 'label')
        option_label.text = str(option.get('label', option.get('name', '')))

        # Explanation (default to empty str)
        option_explanation = etree.SubElement(option_el, 'explanation')
        option_explanation.text = str(option.get('explanation', ''))


def _serialize_criteria(criteria_root, criteria_list):
    """
    Serialize rubric criteria as XML, adding children to the XML
    with root node `criteria_root`.

    We don't make any assumptions about the contents of `criteria_list`,
    and we handle unexpected inputs gracefully.

    Args:
        critera_root (lxml.etree.Element): The root node of the tree.
        criteria_list (list): List of criteria dictionaries.

    Returns:
        None
    """

    # Sort the criteria by order number, then serialize as XML
    for criterion in _sort_by_order_num(criteria_list):
        criterion_el = etree.SubElement(criteria_root, 'criterion')

        # Criterion name (default to a UUID)
        criterion_name = etree.SubElement(criterion_el, 'name')
        if 'name' in criterion:
            criterion_name.text = str(criterion['name'])
        else:
            criterion_name.text = str(uuid().hex)

        # Criterion label (default to the name, then an empty string)
        criterion_label = etree.SubElement(criterion_el, 'label')
        criterion_label.text = str(criterion.get('label', criterion.get('name', '')))

        # Criterion prompt (default to empty string)
        criterion_prompt = etree.SubElement(criterion_el, 'prompt')
        criterion_prompt.text = str(criterion.get('prompt', ''))

        # Criterion feedback disabled, optional, or required
        # If disabled, do not set the attribute.
        if criterion.get('feedback') in ["optional", "required"]:
            criterion_el.set('feedback', criterion['feedback'])

        # Criterion options
        options_list = criterion.get('options', None)
        if isinstance(options_list, list):
            _serialize_options(criterion_el, options_list)


def serialize_rubric(rubric_root, oa_block):
    """
    Serialize a rubric dictionary as XML, adding children to the XML
    with root node `rubric_root`.

    This is very liberal in what it accepts.  If the rubric dict persisted
    by the XBlock is invalid for some reason, we still want to generate XML
    so that Studio authors can fix the error.

    Args:
        oa_block (OpenAssessmentBlock): The OpenAssessmentBlock to serialize
        rubric_dict (dict): A dictionary representation of the rubric, of the form
            described in the serialized Rubric model (peer grading serializers).

    Returns:
        None
    """
    # Criteria
    criteria_list = oa_block.rubric_criteria

    if isinstance(criteria_list, list):
        _serialize_criteria(rubric_root, criteria_list)

    if oa_block.rubric_feedback_prompt is not None:
        feedback_prompt = etree.SubElement(rubric_root, 'feedbackprompt')
        feedback_prompt.text = str(oa_block.rubric_feedback_prompt)

    if oa_block.rubric_feedback_default_text is not None:
        feedback_text = etree.SubElement(rubric_root, 'feedback_default_text')
        feedback_text.text = str(oa_block.rubric_feedback_default_text)


def parse_date(date_str, name=""):
    """
    Attempt to parse a date string into ISO format (without milliseconds)
    Returns `None` if this cannot be done.

    Args:
        date_str (str): The date string to parse.

    Kwargs:
        name (str): the name to return in an error to the origin of the call if an error occurs.

    Returns:
        unicode in ISO format (without milliseconds) if the date string is
        parse-able. None if parsing fails.

    Raises:
        UpdateFromXmlError
    """
    if date_str == "":
        return None
    try:
        # Get the date into ISO format
        parsed_date = dateutil.parser.parse(str(date_str)).replace(tzinfo=pytz.utc)
        formatted_date = parsed_date.strftime("%Y-%m-%dT%H:%M:%S")
        return str(formatted_date)
    except (ValueError, TypeError) as ex:
        msg = (
            'The format of the given date ({date}) for the {name} is invalid. '
            'Make sure the date is formatted as YYYY-MM-DDTHH:MM:SS.'
        ).format(date=date_str, name=name)
        raise UpdateFromXmlError(msg) from ex


def _parse_boolean(boolean_str):
    """
    Attempt to parse a boolean string into a boolean value. Leniently accepts
    both 'True' and 'true', but is otherwise declared false.

    Args:
        boolean_str (unicode): The boolean string to parse.

    Returns:
        The boolean value of the string. True if the string equals 'True' or
        'true'
    """
    return boolean_str in ['True', 'true']


def _parse_prompts_xml(root):
    """
    Parse <prompts> element in the OpenAssessment XBlock's content XML.

    Args:
        root (lxml.etree.Element): The root node of the tree.

    Returns:
        list of prompts dictionaries.

    Raises:
        UpdateFromXmlError: The XML definition is invalid or the XBlock could not be updated.
    """
    prompts_list = []

    prompts_el = root.find('prompts')
    if prompts_el is not None:
        for prompt in prompts_el.findall('prompt'):
            prompt_dict = dict()

            # Prompt description
            prompt_description = prompt.find('description')
            if prompt_description is not None:
                prompt_dict['description'] = _safe_get_text(prompt_description)
            else:
                raise UpdateFromXmlError('Every "prompt" element must contain a "description" element.')

            prompts_list.append(prompt_dict)
    else:
        # For backwards compatibility. Initially a single prompt element was added in
        # the rubric element.
        rubric_el = root.find('rubric')
        prompt_el = rubric_el.find('prompt')
        prompt_description = ''
        if prompt_el is not None:
            prompt_description = _safe_get_text(prompt_el)

        prompts_list.append(
            {
                'description': prompt_description,
            }
        )

    return prompts_list


def _parse_options_xml(options_root):
    """
    Parse <options> element in the OpenAssessment XBlock's content XML.

    Args:
        options_root (lxml.etree.Element): The root of the tree.

    Returns:
        list of option dictionaries, as defined in the Rubric model of the peer grading app.

    Raises:
        UpdateFromXmlError: The XML definition is invalid or the XBlock could not be updated.
    """
    options_list = []
    order_num = 0

    for option in options_root.findall('option'):
        option_dict = dict()

        # Option order number (sequential)
        option_dict['order_num'] = order_num
        order_num += 1

        # Option points -- must be an integer!
        if 'points' in option.attrib:
            try:
                option_dict['points'] = int(option.get('points'))
            except ValueError as ex:
                raise UpdateFromXmlError('The value for "points" must be an integer.') from ex
        else:
            raise UpdateFromXmlError('Every "option" element must contain a "points" attribute.')

        # Option name
        option_name = option.find('name')
        if option_name is not None:
            option_dict['name'] = _safe_get_text(option_name)
        else:
            raise UpdateFromXmlError('Every "option" element must contain a "name" element.')

        # Option label
        # Backwards compatibility: Older problem definitions won't have this.
        # If no label is defined, default to the option name.
        option_label = option.find('label')
        option_dict['label'] = (
            _safe_get_text(option_label)
            if option_label is not None
            else option_dict['name']
        )

        # Option explanation
        option_explanation = option.find('explanation')
        if option_explanation is not None:
            option_dict['explanation'] = _safe_get_text(option_explanation)
        else:
            raise UpdateFromXmlError('Every "option" element must contain an "explanation" element.')

        # Add the options dictionary to the list
        options_list.append(option_dict)

    return options_list


def _parse_criteria_xml(criteria_root):
    """
    Parse <criteria> element in the OpenAssessment XBlock's content XML.

    Args:
        criteria_root (lxml.etree.Element): The root node of the tree.

    Returns:
        list of criteria dictionaries, as defined in the Rubric model of the peer grading app.

    Raises:
        UpdateFromXmlError: The XML definition is invalid or the XBlock could not be updated.
    """
    criteria_list = []
    order_num = 0

    for criterion in criteria_root.findall('criterion'):
        criterion_dict = dict()

        # Criterion order number (sequential)
        criterion_dict['order_num'] = order_num
        order_num += 1

        # Criterion name
        criterion_name = criterion.find('name')
        if criterion_name is not None:
            criterion_dict['name'] = _safe_get_text(criterion_name)
        else:
            raise UpdateFromXmlError('Every "criterion" element must contain a "name" element.')

        # Criterion label
        # Backwards compatibility: Older problem definitions won't have this,
        # so if it isn't set, default to the criterion name.
        criterion_label = criterion.find('label')
        criterion_dict['label'] = (
            _safe_get_text(criterion_label)
            if criterion_label is not None
            else criterion_dict['name']
        )

        # Criterion prompt
        criterion_prompt = criterion.find('prompt')
        if criterion_prompt is not None:
            criterion_dict['prompt'] = _safe_get_text(criterion_prompt)
        else:
            raise UpdateFromXmlError('Every "criterion" element must contain a "prompt" element.')

        # Criterion feedback (disabled, optional, or required)
        criterion_feedback = criterion.get('feedback', 'disabled')
        if criterion_feedback in ['optional', 'disabled', 'required']:
            criterion_dict['feedback'] = criterion_feedback
        else:
            raise UpdateFromXmlError(
                'Invalid value for "feedback" attribute: if specified, it must be set set to "optional" or "required".'
            )

        # Criterion options
        criterion_dict['options'] = _parse_options_xml(criterion)

        # Add the newly constructed criterion dict to the list
        criteria_list.append(criterion_dict)

    return criteria_list


def parse_rubric_xml(rubric_root):
    """
    Parse <rubric> element in the OpenAssessment XBlock's content XML.

    Args:
        rubric_root (lxml.etree.Element): The root of the <rubric> node in the tree.

    Returns:
        dict, a serialized representation of a rubric, as defined by the peer grading serializers.

    Raises:
        UpdateFromXmlError: The XML definition is invalid or the XBlock could not be updated.
        InvalidRubricError: The rubric was not semantically valid.
    """
    rubric_dict = dict()

    feedback_prompt_el = rubric_root.find('feedbackprompt')
    if feedback_prompt_el is not None:
        rubric_dict['feedbackprompt'] = _safe_get_text(feedback_prompt_el)
    else:
        rubric_dict['feedbackprompt'] = None

    feedback_text_el = rubric_root.find('feedback_default_text')
    if feedback_text_el is not None:
        rubric_dict['feedback_default_text'] = _safe_get_text(feedback_text_el)
    else:
        rubric_dict['feedback_default_text'] = None

    # Criteria
    rubric_dict['criteria'] = _parse_criteria_xml(rubric_root)

    return rubric_dict


def parse_examples_xml(examples):
    """
    Parse <example> (training examples) from the XML.

    Args:
        examples (list of lxml.etree.Element): The <example> elements to parse.

    Returns:
        list of example dicts

    Raises:
        UpdateFromXmlError

    """
    examples_list = []
    for example_el in examples:
        example_dict = dict()

        # Retrieve the answers from the training example
        answers_list = list()
        answer_elements = example_el.findall('answer')
        if len(answer_elements) != 1:
            raise UpdateFromXmlError('Each "example" element must contain exactly one "answer" element')

        answer_part_elements = answer_elements[0].findall('part')
        if answer_part_elements:
            for answer_part_element in answer_part_elements:
                answers_list.append(_safe_get_text(answer_part_element))
        else:
            # Initially example answers had only one part.
            answers_list.append(_safe_get_text(answer_elements[0]))

        example_dict['answer'] = {"parts": [{"text": text} for text in answers_list]}

        # Retrieve the options selected from the training example
        example_dict['options_selected'] = []
        for select_el in example_el.findall('select'):
            if 'criterion' not in select_el.attrib:
                raise UpdateFromXmlError('Each "select" element must have a "criterion" attribute')
            if 'option' not in select_el.attrib:
                raise UpdateFromXmlError('Each "select" element must have an "option" attribute')

            example_dict['options_selected'].append({
                'criterion': str(select_el.get('criterion')),
                'option': str(select_el.get('option'))
            })

        examples_list.append(example_dict)

    return examples_list


def parse_assessments_xml(assessments_root):
    """
    Parse the <assessments> element in the OpenAssessment XBlock's content XML.

    Args:
        assessments_root (lxml.etree.Element): The root of the <assessments> node in the tree.

    Returns:
        list of assessment dicts

    Raises:
        UpdateFromXmlError

    """
    assessments_list = []

    for assessment in assessments_root.findall('assessment'):

        assessment_dict = dict()

        # Assessment name
        if 'name' in assessment.attrib:
            assessment_dict['name'] = str(assessment.get('name'))
        else:
            raise UpdateFromXmlError('All "assessment" elements must contain a "name" element.')

        # Assessment start
        if 'start' in assessment.attrib:

            # Other assessment types CAN have a start date
            # pylint: disable=unicode-format-string
            parsed_start = parse_date(assessment.get('start'), name="{} start date".format(assessment_dict['name']))

            if parsed_start is not None:
                assessment_dict['start'] = parsed_start
        else:
            assessment_dict['start'] = None

        # Assessment due
        if 'due' in assessment.attrib:

            # Other assessment types CAN have a due date
            # pylint: disable=unicode-format-string
            parsed_due = parse_date(assessment.get('due'), name="{} due date".format(assessment_dict['name']))

            if parsed_due is not None:
                assessment_dict['due'] = parsed_due
        else:
            assessment_dict['due'] = None

        # Assessment must_grade
        if 'must_grade' in assessment.attrib:
            try:
                assessment_dict['must_grade'] = int(assessment.get('must_grade'))
            except ValueError as ex:
                raise UpdateFromXmlError('The "must_grade" value must be a positive integer.') from ex

        # Assessment must_be_graded_by
        if 'must_be_graded_by' in assessment.attrib:
            try:
                assessment_dict['must_be_graded_by'] = int(assessment.get('must_be_graded_by'))
            except ValueError as ex:
                raise UpdateFromXmlError('The "must_be_graded_by" value must be a positive integer.') from ex

        # Assessment enable_flexible_grading
        if 'enable_flexible_grading' in assessment.attrib:
            try:
                assessment_dict['enable_flexible_grading'] = _parse_boolean(assessment.get('enable_flexible_grading'))
            except ValueError as ex:
                raise UpdateFromXmlError('The "enable_flexible_grading" value must be a boolean.') from ex

        # Assessment required
        if 'required' in assessment.attrib:

            # Staff assessment is the only type to use an explicit required marker
            if assessment_dict['name'] != 'staff-assessment':
                raise UpdateFromXmlError('The "required" field is only allowed for staff assessment.')
            assessment_dict['required'] = _parse_boolean(str(assessment.get('required')))

        # Training examples
        examples = assessment.findall('example')

        # Student training and AI Grading should always have examples set, even if it's an empty list.
        # (Validation rules, applied later, are responsible for
        # ensuring that users specify at least one example).
        # All assessments except for Student Training ignore examples.
        if assessment_dict['name'] == 'student-training':
            assessment_dict['examples'] = parse_examples_xml(examples)

        # Update the list of assessments
        assessments_list.append(assessment_dict)

    return assessments_list


def serialize_training_examples(examples, assessment_el):
    """
    Serialize a training example to XML.

    Args:
        examples (list of dict): List of example dictionaries.
        assessment_el (lxml.etree.Element): The <assessment> XML element.

    Returns:
        None

    """
    for example_dict in examples:
        example_el = etree.SubElement(assessment_el, 'example')

        # Answer provided in the example (default to empty string)
        answer_el = etree.SubElement(example_el, 'answer')
        try:
            answer = example_dict.get('answer')
            if answer is None:
                parts = []
            elif isinstance(answer, dict):
                parts = answer.get('parts', [])
            elif isinstance(answer, list):
                parts = answer

            for part in parts:
                part_el = etree.SubElement(answer_el, 'part')
                # pylint: disable=unicode-format-string
                part_el.text = str(part.get('text', ''))
        except Exception:  # excuse the bare-except, looking for more information on EDUCATOR-1817
            log.exception('Error parsing training example: %s', example_dict)
            raise

        # Options selected from the rubric
        options_selected = example_dict.get('options_selected', [])
        for selected_dict in options_selected:
            select_el = etree.SubElement(example_el, 'select')
            select_el.set('criterion', str(selected_dict.get('criterion', '')))
            select_el.set('option', str(selected_dict.get('option', '')))


def serialize_assessments(assessments_root, oa_block):
    """
    Serialize the assessment modules for an OpenAssessment XBlock.

    Args:
        assessments_root (lxml.etree.Element): The <assessments> XML element.
        oa_block (OpenAssessmentXBlock): The XBlock with configuration to
            serialize.

    Returns:
        None

    """
    for assessment_dict in update_assessments_format(oa_block.rubric_assessments):

        assessment = etree.SubElement(assessments_root, 'assessment')

        # Set assessment attributes, defaulting to empty values
        assessment.set('name', str(assessment_dict.get('name', '')))

        if 'must_grade' in assessment_dict:
            assessment.set('must_grade', str(assessment_dict['must_grade']))

        if 'must_be_graded_by' in assessment_dict:
            assessment.set('must_be_graded_by', str(assessment_dict['must_be_graded_by']))

        if 'enable_flexible_grading' in assessment_dict:
            assessment.set('enable_flexible_grading', str(assessment_dict['enable_flexible_grading']))

        if assessment_dict.get('start') is not None:
            assessment.set('start', str(assessment_dict['start']))

        if assessment_dict.get('due') is not None:
            assessment.set('due', str(assessment_dict['due']))

        if assessment_dict.get('required') is not None:
            assessment.set('required', str(assessment_dict['required']))

        # Training examples
        examples = assessment_dict.get('examples', [])
        if not isinstance(examples, list):
            examples = []
        serialize_training_examples(examples, assessment)


def serialize_content_to_xml(oa_block, root):
    """
    Serialize the OpenAssessment XBlock's content to XML.

    Args:
        oa_block (OpenAssessmentBlock): The open assessment block to serialize.
        root (etree.Element): The XML root node to update.

    Returns:
        etree.Element

    """
    root.tag = 'openassessment'

    # Set the submission start date
    if oa_block.submission_start is not None:
        root.set('submission_start', str(oa_block.submission_start))

    # Set submission due date
    if oa_block.submission_due is not None:
        root.set('submission_due', str(oa_block.submission_due))

    # Set leaderboard show
    if oa_block.leaderboard_show:
        root.set('leaderboard_show', str(oa_block.leaderboard_show))

    # Set text response
    if oa_block.text_response:
        root.set('text_response', str(oa_block.text_response))

    # Set text response editor
    if oa_block.text_response_editor:
        root.set('text_response_editor', str(oa_block.text_response_editor))

    # Set file upload response
    if oa_block.file_upload_response:
        root.set('file_upload_response', str(oa_block.file_upload_response))

    # Set File upload settings
    if oa_block.file_upload_type:
        root.set('file_upload_type', str(oa_block.file_upload_type))

    # Set File type white listing
    if oa_block.white_listed_file_types:
        root.set('white_listed_file_types', str(oa_block.white_listed_file_types_string))

    if oa_block.allow_multiple_files is not None:
        root.set('allow_multiple_files', str(oa_block.allow_multiple_files))

    if oa_block.allow_latex is not None:
        root.set('allow_latex', str(oa_block.allow_latex))

    # Set group access setting if not empty
    if oa_block.group_access:
        root.set('group_access', json.dumps(GroupAccessDict().to_json(oa_block.group_access)))

    # Open assessment displayed title
    title = etree.SubElement(root, 'title')
    title.text = str(oa_block.title)

    # Assessment list
    assessments_root = etree.SubElement(root, 'assessments')
    serialize_assessments(assessments_root, oa_block)

    # Prompts
    prompts_root = etree.SubElement(root, 'prompts')
    _serialize_prompts(prompts_root, oa_block.prompts)

    root.set('prompts_type', str(oa_block.prompts_type))

    # Rubric
    rubric_root = etree.SubElement(root, 'rubric')
    serialize_rubric(rubric_root, oa_block)

    # Team info
    if oa_block.teams_enabled is not None:
        root.set('teams_enabled', str(oa_block.teams_enabled))
    if oa_block.selected_teamset_id is not None:
        root.set('selected_teamset_id', str(oa_block.selected_teamset_id))

    if oa_block.show_rubric_during_response is not None:
        root.set('show_rubric_during_response', str(oa_block.show_rubric_during_response))


def serialize_content(oa_block):
    """
    Serialize the OpenAssessment XBlock's content to an XML string.

    Args:
        oa_block (OpenAssessmentBlock): The open assessment block to serialize.

    Returns:
        xml (unicode)
    """
    root = etree.Element('openassessment')
    serialize_content_to_xml(oa_block, root)

    # Return a UTF-8 representation of the XML
    return etree.tostring(root, pretty_print=True, encoding='unicode')


def serialize_rubric_to_xml_str(oa_block):
    """
    Serialize the OpenAssessment XBlock's rubric into an XML string. This is
    designed to serialize the XBlock's rubric specifically for authoring. Since
    the authoring view splits the prompt from the rubric, the serialized format
    for the rubric does not contain the prompt.

    Args:
        oa_block (OpenAssessmentBlock): The open assessment block to serialize
            a rubric from.

    Returns:
        xml (unicode) representation of the Rubric.

    """
    rubric_root = etree.Element('rubric')
    serialize_rubric(rubric_root, oa_block)
    return etree.tostring(rubric_root, pretty_print=True, encoding='unicode')


def serialize_examples_to_xml_str(assessment):
    """
    Serializes the OpenAssessment XBlock's training examples into an XML unicode
    string.

    Args:
        assessment (dict): Dictionary representation of an Assessment Module's
            configuration. If this contains a list of examples, the examples
            will be returned serialized.

    Returns:
        A unicode string of the XML serialized examples.

    """
    examples = update_assessments_format([assessment])[0].get('examples', [])
    if not isinstance(examples, list):
        examples = []
    examples_root = etree.Element('examples')
    serialize_training_examples(examples, examples_root)
    return etree.tostring(examples_root, pretty_print=True, encoding='unicode')


def serialize_assessments_to_xml_str(oa_block):
    """
    Serializes the OpenAssessment XBlock's assessment modules into an XML
    unicode string.

    Args:
        oa_block (OpenAssessmentBlock
    """
    assessments_root = etree.Element('assessments')
    serialize_assessments(assessments_root, oa_block)
    return etree.tostring(assessments_root, pretty_print=True, encoding='unicode')


def parse_from_xml(root):
    """
    Update the OpenAssessment XBlock's content from an XML definition.

    We need to be strict about the XML we accept, to avoid setting
    the XBlock to an invalid state (which will then be persisted).

    Args:
        root (lxml.etree.Element): The XML definition of the XBlock's content.

    Returns:
        A dictionary of all of the XBlock's content.

    Raises:
        UpdateFromXmlError: The XML definition is invalid
    """

    # Check that the root has the correct tag
    if root.tag != 'openassessment':
        raise UpdateFromXmlError('Every open assessment problem must contain an "openassessment" element.')

    # Retrieve the start date for the submission
    # Set it to None by default; we will update it to the latest start date later on
    submission_start = None
    if 'submission_start' in root.attrib:
        submission_start = parse_date(str(root.attrib['submission_start']), name="submission start date")

    # Retrieve the due date for the submission
    # Set it to None by default; we will update it to the earliest deadline later on
    submission_due = None
    if 'submission_due' in root.attrib:
        submission_due = parse_date(str(root.attrib['submission_due']), name="submission due date")

    text_response = None
    if 'text_response' in root.attrib:
        text_response = str(root.attrib['text_response'])

    text_response_editor = 'text'
    if 'text_response_editor' in root.attrib:
        text_response_editor = str(root.attrib['text_response_editor'])

    file_upload_response = None
    if 'file_upload_response' in root.attrib:
        file_upload_response = str(root.attrib['file_upload_response'])

    allow_file_upload = None
    if 'allow_file_upload' in root.attrib:
        allow_file_upload = _parse_boolean(str(root.attrib['allow_file_upload']))

    file_upload_type = None
    if 'file_upload_type' in root.attrib:
        file_upload_type = str(root.attrib['file_upload_type'])

    white_listed_file_types = None
    if 'white_listed_file_types' in root.attrib:
        white_listed_file_types = str(root.attrib['white_listed_file_types'])

    allow_multiple_files = True
    if 'allow_multiple_files' in root.attrib:
        allow_multiple_files = _parse_boolean(str(root.attrib['allow_multiple_files']))

    allow_latex = False
    if 'allow_latex' in root.attrib:
        allow_latex = _parse_boolean(str(root.attrib['allow_latex']))

    group_access = {}
    if 'group_access' in root.attrib:
        group_access = GroupAccessDict().from_json(json.loads(root.attrib['group_access']))

    show_rubric_during_response = False
    if 'show_rubric_during_response' in root.attrib:
        show_rubric_during_response = _parse_boolean(str(root.attrib['show_rubric_during_response']))

    # Retrieve the title
    title_el = root.find('title')
    if title_el is None:
        raise UpdateFromXmlError('Every assessment must contain a "title" element.')
    title = _safe_get_text(title_el)

    # Retrieve the rubric
    rubric_el = root.find('rubric')
    if rubric_el is None:
        raise UpdateFromXmlError('Every assessment must contain a "rubric" element.')
    rubric = parse_rubric_xml(rubric_el)

    # Retrieve the prompts
    prompts = _parse_prompts_xml(root)

    prompts_type = 'text'
    if 'prompts_type' in root.attrib:
        prompts_type = str(root.attrib['prompts_type'])

    # Retrieve the leaderboard if it exists, otherwise set it to 0
    leaderboard_show = 0
    if 'leaderboard_show' in root.attrib:
        try:
            leaderboard_show = int(root.attrib['leaderboard_show'])
        except (TypeError, ValueError) as ex:
            raise UpdateFromXmlError('The leaderboard must have an integer value.') from ex

    # Retrieve teams info
    teams_enabled = False
    selected_teamset_id = None
    if 'teams_enabled' in root.attrib:
        teams_enabled = _parse_boolean(str(root.attrib['teams_enabled']))
    if 'selected_teamset_id' in root.attrib:
        selected_teamset_id = str(root.attrib['selected_teamset_id'])

    # Retrieve the assessments
    assessments_el = root.find('assessments')
    if assessments_el is None:
        raise UpdateFromXmlError('Every assessment must contain an "assessments" element.')
    assessments = parse_assessments_xml(assessments_el)

    return {
        'title': title,
        'prompts': prompts,
        'prompts_type': prompts_type,
        'rubric_criteria': rubric['criteria'],
        'rubric_assessments': assessments,
        'rubric_feedback_prompt': rubric['feedbackprompt'],
        'rubric_feedback_default_text': rubric['feedback_default_text'],
        'submission_start': submission_start,
        'submission_due': submission_due,
        'text_response': text_response,
        'text_response_editor': text_response_editor,
        'file_upload_response': file_upload_response,
        'allow_file_upload': allow_file_upload,
        'file_upload_type': file_upload_type,
        'white_listed_file_types': white_listed_file_types,
        'allow_multiple_files': allow_multiple_files,
        'allow_latex': allow_latex,
        'group_access': group_access,
        'leaderboard_show': leaderboard_show,
        'teams_enabled': teams_enabled,
        'selected_teamset_id': selected_teamset_id,
        'show_rubric_during_response': show_rubric_during_response,
    }


def parse_from_xml_str(xml):
    """
    Create a dictionary for the OpenAssessment XBlock's content from an XML
    string definition. Parses the string using a library that avoids some known
    security vulnerabilities in etree.

    Args:
        xml (unicode): The XML definition of the XBlock's content.

    Returns:
        A dictionary of all configuration values for the XBlock.

    Raises:
        UpdateFromXmlError: The XML definition is invalid.
        InvalidRubricError: The rubric was not semantically valid.
        InvalidAssessmentsError: The assessments are not semantically valid.
    """
    return parse_from_xml(_unicode_to_xml(xml))


def _unicode_to_xml(xml):
    """
    Converts unicode string to XML node.

    Args:
        xml (unicode): The XML definition of some XBlock configuration.

    Raises:
        UpdateFromXmlError: Raised when the XML definition is invalid.

    """
    # Parse the XML content definition
    # Use the defusedxml library implementation to avoid known security vulnerabilities in ElementTree:
    # http://docs.python.org/2/library/xml.html#xml-vulnerabilities
    try:
        return safe_etree.fromstring(xml.encode('utf-8'))
    except (ValueError, safe_etree.ParseError) as ex:
        raise UpdateFromXmlError("An error occurred while parsing the XML content.") from ex


def parse_examples_from_xml_str(xml):
    """
    Converts an XML string of examples (Student Training or AI) into a dictionary
    representing the same information.

    Args:
        xml (unicode): The XML definition of the examples

    Returns
        (list of dict): The example definition
    """
    examples_root = _unicode_to_xml(xml)
    examples = examples_root.findall('example')
    return parse_examples_xml(examples)



"""
Data Conversion utility methods for handling assessment data transformations.

"""


def update_training_example_answer_format(answer):
    """
    For each training example update 'answer' to newer format.

    Args:
        answer unicode string or dict
    Returns:
        dict
    """
    if isinstance(answer, str):
        return {
            'parts': [
                {'text': answer}
            ]
        }

    return answer



"""
Fields and methods used by the LMS and Studio.
"""
from xblock.fields import DateTime, Dict, Float, Scope, String


class GroupAccessDict(Dict):
    """Special Dict class for serializing the group_access field"""
    def from_json(self, access_dict):  # pylint: disable=arguments-differ
        if access_dict is not None:
            return {int(k): access_dict[k] for k in access_dict}
        return None

    def to_json(self, access_dict):  # pylint: disable=arguments-differ
        if access_dict is not None:
            return {str(k): access_dict[k] for k in access_dict}
        return None


class LmsCompatibilityMixin:
    """
    Extra fields and methods used by LMS/Studio.
    """
    # Studio the default value for this field to show this XBlock
    # in the list of "Advanced Components"
    display_name = String(
        default="Open Response Assessment", scope=Scope.settings,
        help="Display name"
    )

    start = DateTime(
        default=None, scope=Scope.settings,
        help="ISO-8601 formatted string representing the start date of this assignment."
    )

    due = DateTime(
        default=None, scope=Scope.settings,
        help="ISO-8601 formatted string representing the due date of this assignment."
    )

    weight = Float(
        display_name="Problem Weight",
        help=("Defines the number of points each problem is worth. "
              "If the value is not set, the problem is worth the sum of the "
              "option point values."),
        values={"min": 0, "step": .1},
        scope=Scope.settings
    )

    group_access = GroupAccessDict(
        help=(
            "A dictionary that maps which groups can be shown this block. The keys "
            "are group configuration ids and the values are a list of group IDs. "
            "If there is no key for a group configuration or if the set of group IDs "
            "is empty then the block is considered visible to all. Note that this "
            "field is ignored if the block is visible_to_staff_only."
        ),
        default={},
        scope=Scope.settings,
    )

    icon_class = "problem"

    def has_dynamic_children(self):
        """Do we dynamically determine our children? No, we don't have any.

        The LMS wants to know this to see if it has to instantiate our module
        and query it to find the children, or whether it can just trust what's
        in the static (cheaper) children listing.
        """
        return False

    @property
    def has_score(self):
        """Are we a scored type (read: a problem). Yes.

        For LMS Progress page/grades download purposes, we're always going to
        have a score, even if it's just 0 at the start.
        """
        return True

    def max_score(self):
        """The maximum raw score of our problem.

        Called whenever the LMS knows that something is scorable, but finds no
        recorded raw score for it (i.e. the student hasn't done it). In that
        case, the LMS knows that the earned score is 0, but it doesn't know what
        to put in the denominator. So we supply it with the total number of
        points that it is possible for us to earn -- the sum of the highest
        pointed options from each criterion.

        Note that if we have already recorded a score in submissions, this
        method will never be called. So it's perfectly possible for us to have
        10/10 on the progress page and a 12 returning from this method if our
        10/10 score was earned in the past and the problem has changed since
        then.
        """
        return sum(
            max(option["points"] for option in criterion["options"])
            if criterion["options"] else 0
            for criterion in self.rubric_criteria
        )



"""
Aggregate data for openassessment.
"""

from collections import defaultdict, namedtuple, OrderedDict
from io import StringIO
from urllib.parse import urljoin
from zipfile import ZipFile
from itertools import chain
import csv
import json
import os

import requests

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import F
from django.utils.translation import ugettext as _

from submissions import api as sub_api
from submissions.errors import SubmissionNotFoundError
from openassessment.assessment.models import Assessment, AssessmentFeedback, AssessmentPart
from openassessment.fileupload.api import get_download_url
from openassessment.workflow.models import AssessmentWorkflow, TeamAssessmentWorkflow


class CsvWriter:
    """
    Dump openassessment data to CSV files.
    """

    MODELS = [
        'assessment', 'assessment_part',
        'assessment_feedback', 'assessment_feedback_option',
        'submission', 'score'
    ]

    HEADERS = {
        'assessment': [
            'id', 'submission_uuid', 'scored_at',
            'scorer_id', 'score_type',
            'points_possible', 'feedback',
        ],
        'assessment_part': [
            'assessment_id', 'points_earned',
            'criterion_name', 'criterion_label',
            'option_name', 'option_label', 'feedback'
        ],
        'assessment_feedback': [
            'submission_uuid', 'feedback_text', 'options'
        ],
        'assessment_feedback_option': [
            'id', 'text'
        ],
        'submission': [
            'uuid', 'student_id', 'item_id',
            'submitted_at', 'created_at', 'raw_answer'
        ],
        'score': [
            'submission_uuid',
            'points_earned', 'points_possible',
            'created_at',
        ]
    }

    # Number of submissions to retrieve at a time
    # from the database.  We need to do this in order
    # to avoid loading thousands of records into memory at once.
    QUERY_INTERVAL = 100

    def __init__(self, output_streams, progress_callback=None):
        """
        Configure where the writer will write data.

        You can provide open file handles for each of the available
        models (see `AssessmentCsvWriter.MODELS`).  If you don't
        provide an output stream, the writer won't produce data
        for that model.

        Args:
            output_streams (dictionary): Provide the file handles
                to write CSV data to.

        Keyword Arguments:
            progress_callback (callable): Callable that accepts
                no arguments.  Called once per submission loaded
                from the database.

        Example usage:
            >>> output_streams = {
            >>>     "submission": open('submissions.csv', 'w'),
            >>>     "score": open('scores.csv', 'w')
            >>> }
            >>> writer = AssessmentsCsvWriter(output_streams)
            >>> writer.write_to_csv()

        """
        self.writers = {
            key: csv.writer(file_handle)
            for key, file_handle in output_streams.items()
            if key in self.MODELS
        }
        self._progress_callback = progress_callback

    def write_to_csv(self, course_id):
        """
        Write assessment and submission data for a course to CSV files.

        NOTE: The current implementation optimizes for memory usage,
        but not for the number of database queries.  All the queries
        use indexed fields (the submission uuid), so they should be
        relatively quick.

        Args:
            course_id (unicode): The course ID from which to pull data.

        Returns:
            None

        """
        self._write_csv_headers()

        rubric_points_cache = dict()
        feedback_option_set = set()
        for submission_uuid in self._submission_uuids(course_id):
            self._write_submission_to_csv(submission_uuid)

            # Django 1.4 doesn't follow reverse relations when using select_related,
            # so we select AssessmentPart and follow the foreign key to the Assessment.
            parts = self._use_read_replica(
                AssessmentPart.objects.select_related('assessment', 'option', 'option__criterion')
                .filter(assessment__submission_uuid=submission_uuid)
                .order_by('assessment__pk')
            )
            self._write_assessment_to_csv(parts, rubric_points_cache)

            feedback_query = self._use_read_replica(
                AssessmentFeedback.objects
                .filter(submission_uuid=submission_uuid)
                .prefetch_related('options')
            )
            for assessment_feedback in feedback_query:
                self._write_assessment_feedback_to_csv(assessment_feedback)
                feedback_option_set.update({
                    option for option in assessment_feedback.options.all()  # pylint: disable=unnecessary-comprehension
                })

            if self._progress_callback is not None:
                self._progress_callback()

        # The set of available options should be relatively small,
        # since they're not (currently) user-defined.
        self._write_feedback_options_to_csv(feedback_option_set)

    def _submission_uuids(self, course_id):
        """
        Iterate over submission uuids.
        Makes database calls every N submissions to avoid loading
        all submission uuids into memory at once.

        Args:
            course_id (unicode): The ID of the course to retrieve submissions from.

        Yields:
            submission_uuid (unicode)

        """
        num_results = 0
        start = 0
        total_results = self._use_read_replica(
            AssessmentWorkflow.objects.filter(course_id=course_id)
        ).count()

        while num_results < total_results:
            # Load a subset of the submission UUIDs
            # We're assuming that peer workflows are immutable,
            # so if we counted N at the start of the loop,
            # there should be >= N for us to process.
            end = start + self.QUERY_INTERVAL
            query = self._use_read_replica(
                AssessmentWorkflow.objects
                .filter(course_id=course_id)
                .order_by('created')
            ).values('submission_uuid')[start:end]

            for workflow_dict in query:
                num_results += 1
                yield workflow_dict['submission_uuid']

            start += self.QUERY_INTERVAL

    def _write_csv_headers(self):
        """
        Write the headers (first row) for each output stream.
        """
        for name, writer in self.writers.items():
            writer.writerow(self.HEADERS[name])

    def _write_submission_to_csv(self, submission_uuid):
        """
        Write submission data to CSV.

        Args:
            submission_uuid (unicode): The UUID of the submission to write.

        Returns:
            None

        """
        submission = sub_api.get_submission_and_student(submission_uuid, read_replica=True)
        self._write_unicode('submission', [
            submission['uuid'],
            submission['student_item']['student_id'],
            submission['student_item']['item_id'],
            submission['submitted_at'],
            submission['created_at'],
            json.dumps(submission['answer'])
        ])

        score = sub_api.get_latest_score_for_submission(submission_uuid, read_replica=True)
        if score is not None:
            self._write_unicode('score', [
                score['submission_uuid'],
                score['points_earned'],
                score['points_possible'],
                score['created_at']
            ])

    def _write_assessment_to_csv(self, assessment_parts, rubric_points_cache):
        """
        Write assessments and assessment parts to CSV.

        Args:
            assessment_parts (list of AssessmentPart): The assessment parts to write,
                not necessarily from the same assessment.
            rubric_points_cache (dict): in-memory cache of points possible by rubric ID.

        Returns:
            None

        """
        assessment_id_set = set()

        for part in assessment_parts:
            self._write_unicode('assessment_part', [
                part.assessment.id,
                part.points_earned,
                part.criterion.name,
                part.criterion.label,
                part.option.name if part.option is not None else "",
                part.option.label if part.option is not None else "",
                part.feedback
            ])

            # If we haven't seen this assessment before, write it
            if part.assessment.id not in assessment_id_set:
                assessment = part.assessment

                # The points possible in the rubric will be the same for
                # every assessment that shares a rubric.  To avoid querying
                # the rubric criteria/options each time, we cache points possible
                # for each rubric ID.
                if assessment.rubric_id in rubric_points_cache:
                    points_possible = rubric_points_cache[assessment.rubric_id]
                else:
                    points_possible = assessment.points_possible
                    rubric_points_cache[assessment.rubric_id] = points_possible

                self._write_unicode('assessment', [
                    assessment.id,
                    assessment.submission_uuid,
                    assessment.scored_at,
                    assessment.scorer_id,
                    assessment.score_type,
                    points_possible,
                    assessment.feedback
                ])
                assessment_id_set.add(assessment.id)

    def _write_assessment_feedback_to_csv(self, assessment_feedback):
        """
        Write feedback on assessments to CSV.

        Args:
            assessment_feedback (AssessmentFeedback): The feedback model to write.

        Returns:
            None

        """
        options_string = ",".join([
            str(option.id) for option in assessment_feedback.options.all()
        ])

        self._write_unicode('assessment_feedback', [
            assessment_feedback.submission_uuid,
            assessment_feedback.feedback_text,
            options_string
        ])

    def _write_feedback_options_to_csv(self, feedback_options):
        """
        Write feedback on assessment options to CSV.

        Args:
            feedback_options (iterable of AssessmentFeedbackOption)

        Returns:
            None

        """
        for option in feedback_options:
            self._write_unicode(
                'assessment_feedback_option',
                [option.id, option.text]
            )

    def _write_unicode(self, output_name, row):
        """
        Encode a row as a UTF-8 bytestring, then write it to a CSV file.
        Non-string values are first converted to unicode.

        Args:
            output_name (str): The name of the output stream to write to.
            row (list): List of fields, which must be serializable as UTF-8.

        Returns:
            None

        """
        writer = self.writers.get(output_name)
        if writer is not None:
            encoded_row = [str(field) for field in row]
            writer.writerow(encoded_row)

    def _use_read_replica(self, queryset):
        """
        Use the read replica if it's available.

        Args:
            queryset (QuerySet)

        Returns:
            QuerySet

        """
        return (
            queryset.using("read_replica")
            if "read_replica" in settings.DATABASES
            else queryset
        )


class OraAggregateData:
    """
    Aggregate all the ORA data into a single table-like data structure.
    """

    @classmethod
    def _use_read_replica(cls, queryset):
        """
        If there's a read replica that can be used, return a cursor to that.
        Otherwise, return a cursor to the regular database.

        Args:
            queryset (QuerySet): The queryset that we would like to use the read replica for.
        Returns:
            QuerySet
        """
        return (
            queryset.using("read_replica")
            if "read_replica" in settings.DATABASES
            else queryset
        )

    @classmethod
    def _usernames_enabled(cls):
        """
        Checks if toggle for deanonymized usernames in report enabled.
        """

        return settings.FEATURES.get('ENABLE_ORA_USERNAMES_ON_DATA_EXPORT', False)

    @classmethod
    def _map_anonymized_ids_to_usernames(cls, anonymized_ids):
        """
        Args:
            anonymized_ids - list of anonymized user ids.
        Returns:
            dictionary, that contains mapping between anonymized user ids and
            actual usernames.
        """
        User = get_user_model()

        users = cls._use_read_replica(
            User.objects.filter(anonymoususerid__anonymous_user_id__in=anonymized_ids)
            .annotate(anonymous_id=F("anonymoususerid__anonymous_user_id"))
            .values("username", "anonymous_id")
        )

        anonymous_id_to_username_mapping = {
            user["anonymous_id"]: user["username"] for user in users
        }

        return anonymous_id_to_username_mapping

    @classmethod
    def _map_sudents_and_scorers_ids_to_usernames(cls, all_submission_information):
        """
        Args:
            all_submission_information - list of tuples with submission data,
            that returned by submissions api's
            `get_all_course_submission_information` method.
        Returns:
            dictionary, that contains mapping between students and scoreres
            anonymized user ids and actual usernames.
        """

        student_ids = []
        submission_uuids = []

        for student_item, submission, _ in all_submission_information:
            student_ids.append(student_item["student_id"])
            submission_uuids.append(submission["uuid"])

        scorer_ids = cls._use_read_replica(
            Assessment.objects.filter(submission_uuid__in=submission_uuids).values_list(
                "scorer_id", flat=True
            )
        )

        return cls._map_anonymized_ids_to_usernames(student_ids + list(scorer_ids))

    @classmethod
    def _map_block_usage_keys_to_display_names(cls, course_id):
        """
        Fetches all course blocks and build mapping between block usage key
        string and block display name for those ones, whoose category is equal
        to ``openassessment``.

        Args:
            course_id (string or CourseLocator instance) - id of course
            resourse
        Returns:
            dictionary, that contains mapping between block usage
            keys (locations) and block display names.
        """
        # pylint: disable=import-error

        from lms.djangoapps.course_blocks.api import get_course_blocks
        from openedx.core.djangoapps.content.block_structure.transformers import BlockStructureTransformers

        from xmodule.modulestore.django import modulestore

        store = modulestore()
        course_usage_key = store.make_course_usage_key(course_id)

        # Passing an empty block structure transformer here to avoid user access checks
        blocks = get_course_blocks(None, course_usage_key, BlockStructureTransformers())

        block_display_name_map = {}

        for block_key in blocks:
            block_type = blocks.get_xblock_field(block_key, 'category')
            if block_type == 'openassessment':
                block_display_name_map[str(block_key)] = blocks.get_xblock_field(block_key, 'display_name')

        return block_display_name_map

    @classmethod
    def _build_assessments_cell(cls, assessments, usernames_map):
        """
        Args:
            assessments (QuerySet) - assessments that we would like to collate into one column.
            usernames_map - dictionary that maps anonymous ids to usernames.
        Returns:
            string that should be included in the 'assessments' column for this set of assessments' row
        """
        usernames_enabled = cls._usernames_enabled()

        returned_string = ""
        for assessment in assessments:
            returned_string += f"Assessment #{assessment.id}\n"
            returned_string += f"-- scored_at: {assessment.scored_at}\n"
            returned_string += f"-- type: {assessment.score_type}\n"
            if usernames_enabled:
                returned_string += "-- scorer_username: {}\n".format(usernames_map.get(assessment.scorer_id, ''))
            returned_string += f"-- scorer_id: {assessment.scorer_id}\n"
            if assessment.feedback != "":
                returned_string += f"-- overall_feedback: {assessment.feedback}\n"
        return returned_string

    @classmethod
    def _build_assessments_parts_cell(cls, assessments):
        """
        Args:
            assessments (QuerySet) - assessments containing the parts that we would like to collate into one column.
        Returns:
            string that should be included in the relevant 'assessments_parts' column for this set of assessments' row
        """
        returned_string = ""
        for assessment in assessments:
            returned_string += f"Assessment #{assessment.id}\n"
            for part in assessment.parts.order_by('criterion__order_num'):
                returned_string += f"-- {part.criterion.label}"
                if part.option is not None and part.option.label is not None:
                    option_label = part.option.label
                    returned_string += ": {option_label} ({option_points})\n".format(
                        option_label=option_label, option_points=part.option.points
                    )
                if part.feedback != "":
                    returned_string += f"-- feedback: {part.feedback}\n"
        return returned_string

    @classmethod
    def _build_assessment_parts_array(cls, assessment, median_scores):
        """
        Args:
            assessment - assessment containing the parts that we would like to report on.
            median_scores - dictionary with criterion name keys and median score values,
               as returned by Assessment.get_median_score_dict()

        Returns:
            OrderedDict that contains an entries for each criterion of the assessment(s).
        """
        parts = OrderedDict()
        number = 1
        for part in assessment.parts.order_by('criterion__order_num'):
            option_label = None
            option_points = None
            if part.option:
                option_label = part.option.label
                option_points = part.option.points

            criterion_col_label = _(u'Criterion {number}: {label}').format(number=number, label=part.criterion.label)
            parts[criterion_col_label] = option_label or ''
            parts[_('Points {number}').format(number=number)] = option_points or 0
            parts[_('Median Score {number}').format(number=number)] = median_scores.get(part.criterion.name)
            parts[_('Feedback {number}').format(number=number)] = part.feedback or ''
            number += 1
        return parts

    @classmethod
    def _build_feedback_options_cell(cls, assessments):
        """
        Args:
            assessments (QuerySet) - assessment that we would like to use to fetch and read the feedback options.
        Returns:
            string that should be included in the relevant 'feedback_options' column for this set of assessments' row
        """

        returned_string = ""
        for assessment in assessments:
            for feedback in assessment.assessment_feedback.all():
                for option in feedback.options.all():
                    returned_string += option.text + "\n"

        return returned_string

    @classmethod
    def _build_feedback_cell(cls, submission_uuid):
        """
        Args:
            submission_uuid (string) - the submission_uuid associated with this particular assessment feedback
        Returns:
            string that should be included in the relevant 'feedback' column for this set of assessments' row
        """
        try:
            feedback = AssessmentFeedback.objects.get(submission_uuid=submission_uuid)
        except AssessmentFeedback.DoesNotExist:
            return ""
        return feedback.feedback_text

    @classmethod
    def _build_response_file_links(cls, submission):
        """
        Args:
            submission - object
        Returns:
            string that contains newline-separated URLs to each of the files uploaded for this submission.
        """
        file_links = ''
        sep = "\n"
        base_url = getattr(settings, 'LMS_ROOT_URL', '')

        from openassessment.xblock.openassessmentblock import OpenAssessmentBlock
        file_downloads = OpenAssessmentBlock.get_download_urls_from_submission(submission)
        for url, _description, _filename, _show_delete in file_downloads:
            if file_links:
                file_links += sep
            file_links += urljoin(base_url, url)
        return file_links

    @classmethod
    def collect_ora2_data(cls, course_id):
        """
        Query database for aggregated ora2 response data.

        Args:
            course_id (string) - the course id of the course whose data we would like to return

        Returns:
            A tuple containing two lists: headers and data.

            headers is a list containing strings corresponding to the column headers of the data.
            data is a list of lists, where each sub-list corresponds to a row in the table of all the data
                for this course.

        """
        all_submission_information = list(sub_api.get_all_course_submission_information(course_id, 'openassessment'))
        usernames_enabled = cls._usernames_enabled()

        usernames_map = (
            cls._map_sudents_and_scorers_ids_to_usernames(all_submission_information)
            if usernames_enabled
            else {}
        )
        block_display_names_map = cls._map_block_usage_keys_to_display_names(course_id)

        rows = []
        for student_item, submission, score in all_submission_information:
            assessments = cls._use_read_replica(
                Assessment.objects.prefetch_related('parts').
                prefetch_related('rubric').
                filter(
                    submission_uuid=submission['uuid']
                )
            )
            assessments_cell = cls._build_assessments_cell(assessments, usernames_map)
            assessments_parts_cell = cls._build_assessments_parts_cell(assessments)
            feedback_options_cell = cls._build_feedback_options_cell(assessments)
            feedback_cell = cls._build_feedback_cell(submission['uuid'])

            row_username_cell = (
                [usernames_map.get(student_item["student_id"], "")]
                if usernames_enabled
                else []
            )

            problem_name = block_display_names_map.get(student_item['item_id'])

            row = [
                submission['uuid'],
                student_item['item_id'],
                problem_name,
                submission['student_item'],
            ] + row_username_cell + [
                student_item['student_id'],
                submission['submitted_at'],
                #  Dumping required to render special characters in CSV
                json.dumps(submission['answer'], ensure_ascii=False),
                assessments_cell,
                assessments_parts_cell,
                score.get('created_at', ''),
                score.get('points_earned', ''),
                score.get('points_possible', ''),
                feedback_options_cell,
                feedback_cell
            ]
            rows.append(row)

        header_username_cell = (
            ['Username']
            if usernames_enabled
            else []
        )

        header = [
            'Submission ID',
            'Location',
            'Problem Name',
            'Item ID'
        ] + header_username_cell + [
            'Anonymized Student ID',
            'Date/Time Response Submitted',
            'Response',
            'Assessment Details',
            'Assessment Scores',
            'Date/Time Final Score Given',
            'Final Score Points Earned',
            'Final Score Points Possible',
            'Feedback Statements Selected',
            'Feedback on Peer Assessments'
        ]
        return header, rows

    @classmethod
    def collect_ora2_summary(cls, course_id):
        """
        Query database for aggregated ora2 summary data.

        Args:
            course_id (string) - the course id of the course whose data we would like to return

        Returns:
            A tuple containing two lists: headers and data.

            headers is a list containing strings corresponding to the column headers of the data.
            data is a list of lists, where each sub-list corresponds to a row in the table of all the data
                for this course.

            Headers details:

            block_name: id of ora block
            student_id: anonymized student id
            status: string indicating the current step or status the student is
                at. Eg. 'peer', 'done', 'cancelled'. Values are from the AssessmentWorkflow
                STEPS + STATUSES
            is_<STEP>_complete: boolean 'complete' status for STEP (0 or 1, or
                empty if workflow does not include this step)
            is_<STEP>_graded: boolean 'graded' status for STEP (0 or 1, or
                empty if workflow does not include this step)
            num_peers_graded: number of peers that 'student_id' has graded in the peer step
            num_graded_by_peers: number of peer grades that 'student_id' has received in the peer step
            is_staff_grade_received: boolean (0 or 1)
            is_final_grade_received: boolean (0 or 1)
            final_grade_points_earned: number of points earned in final grade.
                will be empty if no final grade yet
            final_grade_points_possible: max number of points possible for
                final grade. will be empty if no final grade
        """

        items = AssessmentWorkflow.objects.filter(course_id=course_id)

        # need the workflow steps set and sorted here so the data columns line
        # up with the headers
        steps = sorted(AssessmentWorkflow.STEPS)

        rows = []
        for aw in items:
            statuses = aw.status_details()
            try:
                submission_dict = sub_api.get_submission_and_student(aw.submission_uuid)
            except SubmissionNotFoundError:
                continue

            steps_statuses = []
            peers_graded = 0
            graded_by_count = 0
            for step in steps:
                if not statuses.get(step):
                    # if no status for step, then the 'complete' and 'graded'
                    # statuses should be empty.
                    steps_statuses.append('')
                    steps_statuses.append('')
                    continue

                # if we get to here, then a status exists for `step`

                if statuses[step]['complete']:
                    steps_statuses.append(1)
                else:
                    steps_statuses.append(0)

                if statuses[step]['graded']:
                    steps_statuses.append(1)
                else:
                    steps_statuses.append(0)

                # the peer step is special and has extra metadata
                if step == 'peer':
                    peers_graded = statuses[step]['peers_graded_count'] or 0
                    graded_by_count = statuses[step]['graded_by_count'] or 0

            is_staff_grade_received = 1 if aw.staff_score_exists() else 0
            is_final_grade_received = 1 if aw.status == AssessmentWorkflow.STATUS.done else 0

            score = aw.score
            if score is not None:
                final_grade_points_earned = score['points_earned']
                final_grade_points_possible = score['points_possible']
            else:
                final_grade_points_earned = ''
                final_grade_points_possible = ''

            row = [
                aw.item_id,
                submission_dict['student_item']['student_id'],
                aw.status,
            ] + steps_statuses + [
                peers_graded,
                graded_by_count,
                is_staff_grade_received,
                is_final_grade_received,
                final_grade_points_earned,
                final_grade_points_possible,
            ]
            rows.append(row)

        steps_headers = list(chain.from_iterable((
            (
                "is_{}_complete".format(step),
                "is_{}_graded".format(step),
            )
            for step in steps
        )))

        header = [
            'block_name',
            'student_id',
            'status',
        ] + steps_headers + [
            'num_peers_graded',
            'num_graded_by_peers',
            'is_staff_grade_received',
            'is_final_grade_received',
            'final_grade_points_earned',
            'final_grade_points_possible',
        ]

        return header, rows

    @classmethod
    def collect_ora2_responses(cls, course_id, desired_statuses=None):
        """
        Get information about all ora2 blocks in the course with response count for each step

        Args:
            course_id (string) - the course id of the course whose data we would like to return
            desired_statuses (list) - statuses to return in the result dict for each ora item

        Returns:
            A dict in the format:

            {
             'block-v1:test-org+cs101+2017_TEST+type@openassessment+block@fb668396b505470e914bad8b3178e9e7:
                 {'training': 0, 'self': 0, 'done': 2, 'peer': 1, 'staff': 0, 'total': 3},
             'block-v1:test-org+cs101+2017_TEST+type@openassessment+block@90b4edff50bc47d9ba037a3180c44e97:
                 {'training': 0, 'self': 2, 'done': 0, 'peer': 0, 'staff': 2, 'total': 4},
             ...
            }

        """

        all_valid_ora_statuses = set()
        all_valid_ora_statuses.update(AssessmentWorkflow().STATUS_VALUES)
        all_valid_ora_statuses.update(TeamAssessmentWorkflow().STATUS_VALUES)

        if desired_statuses:
            statuses = [st for st in all_valid_ora_statuses if st in desired_statuses]
        else:
            statuses = all_valid_ora_statuses

        items = AssessmentWorkflow.objects.filter(course_id=course_id, status__in=statuses).values('item_id', 'status')

        result = defaultdict(lambda: {status: 0 for status in statuses})
        for item in items:
            item_id = item['item_id']
            status = item['status']
            result[item_id]['total'] = result[item_id].get('total', 0) + 1
            if status in statuses:
                result[item_id][status] += 1

        return result

    @classmethod
    def generate_assessment_data(cls, xblock_id, submission_uuid=None):
        """
        Generates an OrderedDict for each submission and/or assessment for the given user state.

        Arguments:
        * xblock_id: unique identifier for the current XBlock
        * submission_uuid: unique identifier for the submission, or None
        """
        row = OrderedDict()
        row[_('Item ID')] = xblock_id
        row[_('Submission ID')] = submission_uuid or ''

        submission = None
        if submission_uuid:
            submission = sub_api.get_submission_and_student(submission_uuid)

        if not submission:
            # If no submission, just report block Item ID.
            yield row
            return

        student_item = submission['student_item']
        row[_('Anonymized Student ID')] = student_item['student_id']

        assessments = cls._use_read_replica(
            Assessment.objects.prefetch_related('parts').
            prefetch_related('rubric').
            filter(
                submission_uuid=submission['uuid']
            )
        )
        if assessments:
            scores = Assessment.scores_by_criterion(assessments)
            median_scores = Assessment.get_median_score_dict(scores)
        else:
            # If no assessments, just report submission data.
            median_scores = []
            assessments = [None]

        score = sub_api.get_score(student_item) or {}
        feedback_cell = cls._build_feedback_cell(submission_uuid)
        response_files = cls._build_response_file_links(submission)

        for assessment in assessments:
            assessment_row = row.copy()
            if assessment:
                assessment_cells = cls._build_assessment_parts_array(assessment, median_scores)
                feedback_options_cell = cls._build_feedback_options_cell([assessment])

                score_created_at = score.get('created_at', '')
                if score_created_at:
                    score_created_at = score_created_at.strftime('%F %T %Z')

                assessment_row[_('Assessment ID')] = assessment.id
                assessment_row[_('Assessment Scored Date')] = assessment.scored_at.strftime('%F')
                assessment_row[_('Assessment Scored Time')] = assessment.scored_at.strftime('%T %Z')
                assessment_row[_('Assessment Type')] = assessment.score_type
                assessment_row[_('Anonymous Scorer Id')] = assessment.scorer_id
                assessment_row.update(assessment_cells)
                assessment_row[_('Overall Feedback')] = assessment.feedback or ''
                assessment_row[_('Assessment Score Earned')] = assessment.points_earned
                assessment_row[_('Assessment Scored At')] = assessment.scored_at.strftime('%F %T %Z')
                assessment_row[_('Date/Time Final Score Given')] = score_created_at
                assessment_row[_('Final Score Earned')] = score.get('points_earned', '')
                assessment_row[_('Final Score Possible')] = score.get('points_possible', assessment.points_possible)
                assessment_row[_('Feedback Statements Selected')] = feedback_options_cell
                assessment_row[_('Feedback on Assessment')] = feedback_cell

            assessment_row[_('Response Files')] = response_files
            yield assessment_row


class OraDownloadData:
    """
    Helper class, that is used for downloading and compressing data related
    to submissions (attachments, answer texts).
    """

    ATTACHMENT = 'attachment'
    TEXT = 'text'
    DOWNLOADS_CSV_HEADER = (
        'course_id',
        'block_id',
        'student_id',
        'key',
        'name',
        'type',
        'description',
        'size',
        'file_path',
    )

    @classmethod
    def _download_file_by_key(cls, key):
        download_url = urljoin(
            settings.LMS_ROOT_URL, get_download_url(key)
        )

        response = requests.get(download_url)
        response.raise_for_status()

        return response.content

    @classmethod
    def create_zip_with_attachments(cls, file, course_id, submission_files_data):
        """
        Opens given stream as a zip file and writes into it all submission
        attachments and csv with list of all downloads.

        Example of result zip file structure:
        ```
        .
        └── CourseId
            ├── BlockId1
            │   ├── StudentId1
            │   │   ├── attachments
            │   │   │   ├── SomeFile1
            │   │   │   └── SomeFile2
            │   │   ├── part_0.txt
            │   │   └── part_1.txt
            │   └── StudentId2
            │       ├── part_0.txt
            │       └── part_1.txt
            ├── BlockId2
            │   └── StudentId3
            │       ├── attachments
            │       │   └── SomeFile4
            │       └── part_0.txt
            └── downloads.csv
        ```
        """
        csv_output_buffer = StringIO()

        csvwriter = csv.DictWriter(csv_output_buffer, cls.DOWNLOADS_CSV_HEADER, extrasaction='ignore')
        csvwriter.writeheader()

        with ZipFile(file, 'w') as zip_file:
            for file_data in submission_files_data:
                file_content = (
                    cls._download_file_by_key(file_data['key'])
                    if file_data['type'] == cls.ATTACHMENT
                    else file_data['content']
                )

                zip_file.writestr(file_data['file_path'], file_content)
                csvwriter.writerow(file_data)

            downloads_csv_path = os.path.join(str(course_id), 'downloads.csv')

            zip_file.writestr(
                downloads_csv_path,
                csv_output_buffer.getvalue().encode('utf-8')
            )

        file.seek(0)

        return True

    @classmethod
    def collect_ora2_submission_files(cls, course_id):
        """
        Generator, that yields dictionaries with information about submission
        attachment or answer text.
        """

        all_submission_information = sub_api.get_all_course_submission_information(course_id, 'openassessment')

        for student, submission, _ in all_submission_information:
            raw_answer = submission.get('answer', dict())
            answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(raw_answer)
            for uploaded_file in answer.get_file_uploads():
                yield {
                    'type': cls.ATTACHMENT,
                    'course_id': course_id,
                    'block_id': student['item_id'],
                    'student_id': student['student_id'],
                    'key': uploaded_file.key,
                    'name': uploaded_file.name,
                    'description': uploaded_file.description,
                    'size': uploaded_file.size,
                    'file_path': os.path.join(
                        str(course_id),
                        student['item_id'],
                        student['student_id'],
                        'attachments',
                        uploaded_file.name,
                    )
                }

            # collecting submission answer texts
            for index, text_response in enumerate(answer.get_text_responses()):
                file_name = f'part_{index}.txt'

                yield {
                    'type': cls.TEXT,
                    'course_id': course_id,
                    'block_id': student['item_id'],
                    'student_id': student['student_id'],
                    'key': '',
                    'name': file_name,
                    'description': 'Submission text.',
                    'content': text_response,
                    'size': len(text_response),
                    'file_path': os.path.join(
                        str(course_id),
                        student['item_id'],
                        student['student_id'],
                        file_name,
                    )
                }


class SubmissionFileUpload:
    """
    A SubmissionFileUpload represents a file that was uploaded and submitted as a part of an ORA
    submission. It has the following fields:
        - key: The unique key used by the file upload backend to identify the file.
        - name: The filename of the submitted file.
        - description: An uploader-provided description of the submitted file.
        - size: The filesize of the submitted file.

    Due to historical considerations, only the file key is _required_ to exist,
    but all files uploaded after November 25th, 2019, _should_ contain all fields.

    If fields are missing, they will default to the following values:
        - name: key
        - description: SubmissionFileUpload.DEFAULT_DESCRIPTION
        - size: 0

    A SubmissionFileUpload is distinct from any of the data classes in openassessment/fileupload/api.py.
    FileDescriptor is a display-level construct and FileUpload represents a file that has been uploaded
    but not submitted.
    """

    DEFAULT_DESCRIPTION = _("No description provided.")

    def __init__(self, key, name=None, description=None, size=0):
        self.key = key
        self.name = name if name is not None else SubmissionFileUpload.generate_name_from_key(key)
        self.description = description if description is not None else SubmissionFileUpload.DEFAULT_DESCRIPTION
        self.size = size

    @staticmethod
    def generate_name_from_key(key):
        """
        Return the hex representation of the absolute hash of a value.
        Used to generate arbitrary file names for files with no name.
        """
        return format(abs(hash(key)), 'x')


class OraSubmissionAnswerFactory:
    """ A factory class that takes the parsed json raw_answer from a submission and returns an OraSubmissionAnswer """

    @staticmethod
    def parse_submission_raw_answer(raw_answer):
        """
        Currently this function does a basic test and returns a ZippedListSubmissionAnswer or a TextOnlySubmissionAnswer
        In the future if we were to change the way we do submissions, we would check here and return accordingly.
        """
        if TextOnlySubmissionAnswer.matches(raw_answer):
            return TextOnlySubmissionAnswer(raw_answer)
        elif ZippedListSubmissionAnswer.matches(raw_answer):
            return ZippedListSubmissionAnswer(raw_answer)
        else:
            raise VersionNotFoundException(f"No ORA Submission Answer version recognized for {raw_answer}")


class OraSubmissionAnswer:
    """ Abstract interface for ORA Submissions """
    def __init__(self, raw_answer):
        self.raw_answer = raw_answer

    @staticmethod
    def matches(raw_answer):
        """
        Check if the raw answer fits this type of OraSubmissionAnswer
        """
        raise NotImplementedError()

    def get_text_responses(self):
        """
        Get the list of text responses for the submission

        Returns: list of strings
        """
        raise NotImplementedError()

    def get_file_uploads(self, missing_blank=False):
        """
        Get the list of FileUploads for this submission
        """
        raise NotImplementedError()


class TextOnlySubmissionAnswer(OraSubmissionAnswer):

    @staticmethod
    def matches(raw_answer):
        keys = list(raw_answer.keys())
        return len(keys) == 1 and keys == ['parts']

    def __init__(self, submission):
        super().__init__(submission)
        self.text_responses = None

    def get_text_responses(self):
        """
        Parse and cache text responses from the submission
        """
        if self.text_responses is None:
            self.text_responses = [part.get('text') for part in self.raw_answer.get('parts', [])]
        return self.text_responses

    def get_file_uploads(self, missing_blank=False):
        return []


# This namedtuple represents the different shapes different versions of ORA Submissions have taken.
# Below is a table of the dates and commits that introduced each version:
#   Version | Date              | Commit
#   -----------------------------------------------------------------------
#   1       | July      8, 2014 | 42cf870695c3f2ca010abcf4e69a47d34dc56275
#   2       | April    27, 2017 | 7568a7008706db6fee5d3081e455b6687d84d659
#   3       | October  28, 2019 | 9d8b2de0a04c410c3da7d2894ce0eab8bbc9f254
#   4       | November 12, 2019 | 6c062ecc03e9bcc93d3dc78345cf63bcf910c58f
#   5       | November 25, 2019 | e0e56ac6bc054b7cd71c5e10c8cb99592511cac9
#  -------------------------------------------------------------------------

ZippedListsSubmissionVersion = namedtuple(
    'ZippedListsSubmissionVersion',
    ['key', 'description', 'name', 'size']
)

VERSION_1 = ZippedListsSubmissionVersion('file_key', None, None, None)
VERSION_2 = ZippedListsSubmissionVersion('file_keys', 'files_descriptions', None, None)
VERSION_3 = ZippedListsSubmissionVersion('file_keys', 'files_descriptions', 'files_name', None)
VERSION_4 = ZippedListsSubmissionVersion(
    'file_keys', 'files_descriptions', 'files_name', 'files_sizes'
)
VERSION_5 = ZippedListsSubmissionVersion(
    'file_keys', 'files_descriptions', 'files_names', 'files_sizes'
)
ZIPPED_LIST_SUBMISSION_VERSIONS = [
    VERSION_1, VERSION_2, VERSION_3, VERSION_4, VERSION_5
]


class VersionNotFoundException(Exception):
    """ Raised when we are unable to resolve a given submission to a submission version """


class ZippedListSubmissionAnswer(OraSubmissionAnswer):
    """
    Representation of a type of ORA submission where there are multiple lists, each
    representing a field. They are "zipped" together to represent individual files.
    """
    CURRENT_VERSION = 5

    @staticmethod
    def matches(raw_answer):
        try:
            ZippedListSubmissionAnswer.get_version(raw_answer)
        except VersionNotFoundException:
            return False
        else:
            return True

    @staticmethod
    def does_version_match(submission_keys, version):
        """
        Given a ZippedListsSubmissionVersion and a set of keys from a raw_answer,
        returns whether or not the version and keys match.

        Matching means that either the set of keys in the version matches the given set of keys,
        or the version keys plus the key "parts" matches the given set.
        """
        version_keys = {version_key for version_key in version if version_key}
        if version_keys == submission_keys:
            return True
        version_keys.add('parts')
        return version_keys == submission_keys

    @staticmethod
    def get_version(raw_answer):
        """
        Determines the version associated with a submission by working backwards from the most recent
        submission version and checking if the set of keys in the given submission matches the set
        of keys in the version.

        Raises:
            - VersionNotFoundException if the version cannot be determined.
        """
        submission_keys = set(raw_answer.keys())
        for version in reversed(ZIPPED_LIST_SUBMISSION_VERSIONS):
            if ZippedListSubmissionAnswer.does_version_match(submission_keys, version):
                return version
        raise VersionNotFoundException(f"No zipped list version found with keys {submission_keys}")

    def __init__(self, raw_answer):
        """
        Raises:
            - VersionNotFoundException if a version cannot be matched against the given submission
        """
        super().__init__(raw_answer)
        self.text_responses = None
        self.file_uploads = None
        self.version = ZippedListSubmissionAnswer.get_version(raw_answer)

    def get_text_responses(self):
        """
        Parse and cache text responses from the submission
        """
        if self.text_responses is None:
            self.text_responses = [part.get('text') for part in self.raw_answer.get('parts', [])]
        return self.text_responses

    def _index_safe_get(self, i, target_list, default=None):
        """
        Attempts to get item at target_list[i]. If the index is out of bounds, returns default.
        More or less dict.get() but for lists
        """
        try:
            return target_list[i]
        except IndexError:
            return default

    def get_file_uploads(self, missing_blank=False):
        """
        Parse and cache file upload responses from the raw_answer
        """
        default_missing_value = '' if missing_blank else None
        if self.file_uploads is None:
            file_keys = self.raw_answer.get(self.version.key, [])
            # The very earliest version of ora submissions with files only allowed one file, and so is the only
            #  situation in which any of these fields is not a list
            if not isinstance(file_keys, list):
                file_keys = [file_keys]

            files = []
            file_names = self.raw_answer.get(self.version.name, [])
            file_descriptions = self.raw_answer.get(self.version.description, [])
            file_sizes = self.raw_answer.get(self.version.size, [])
            for i, key in enumerate(file_keys):
                name = self._index_safe_get(i, file_names, default_missing_value)
                description = self._index_safe_get(i, file_descriptions, default_missing_value)
                size = self._index_safe_get(i, file_sizes, 0)

                file_upload = SubmissionFileUpload(key, name=name, description=description, size=size)
                files.append(file_upload)
            self.file_uploads = files
        return self.file_uploads



""" File Upload backends. """


from django.conf import settings

from . import django_storage, filesystem, s3, swift


def get_backend():
    # .. setting_name: ORA2_FILEUPLOAD_BACKEND
    # .. setting_default: 's3'
    # .. setting_description: The backend used to upload the ora2 submissions attachments.
    #     The supported values are: 's3', 'filesystem', 'swift' and 'django'.
    backend_setting = getattr(settings, "ORA2_FILEUPLOAD_BACKEND", "s3")
    if backend_setting == "s3":
        return s3.Backend()
    elif backend_setting == "filesystem":
        return filesystem.Backend()
    elif backend_setting == "swift":
        return swift.Backend()
    elif backend_setting == "django":
        return django_storage.Backend()
    else:
        raise ValueError("Invalid ORA2_FILEUPLOAD_BACKEND setting value: %s" % backend_setting)



"""
Public interface for staff grading, used by students/course staff.
"""


import logging

from django.db import DatabaseError, transaction
from django.utils.timezone import now

from submissions import api as submissions_api

from openassessment.assessment.errors import StaffAssessmentInternalError, StaffAssessmentRequestError
from openassessment.assessment.models import Assessment, AssessmentPart, InvalidRubricSelection, StaffWorkflow
from openassessment.assessment.serializers import InvalidRubric, full_assessment_dict, rubric_from_dict


logger = logging.getLogger("openassessment.assessment.api.staff")  # pylint: disable=invalid-name

STAFF_TYPE = "ST"


def submitter_is_finished(submission_uuid, staff_requirements):  # pylint: disable=unused-argument
    """
    Determine if the submitter has finished their requirements for staff
    assessment. Always returns True.

    Args:
        submission_uuid (str): Not used.
        staff_requirements (dict): Not used.

    Returns:
        True

    """
    return True


def assessment_is_finished(submission_uuid, staff_requirements):
    """
    Determine if the staff assessment step of the given submission is completed.
    This checks to see if staff have completed the assessment.

    Args:
        submission_uuid (str): The UUID of the submission being graded.
        staff_requirements (dict): Any variables that may effect this state.

    Returns:
        True if a staff assessment has been completed for this submission or if not required.
    """
    # Requirements of None means we can't make any assumptions about the done-ness of this step
    if staff_requirements is None:
        return False

    if staff_requirements.get('required', False):
        return bool(get_latest_staff_assessment(submission_uuid))

    return True


def on_init(submission_uuid):
    """
    Create a new staff workflow for a student item and submission.

    Creates a unique staff workflow for a student item, associated with a
    submission.

    Note that the staff workflow begins things in on_init() instead of
    on_start(), because staff shoud be able to access the submission
    regardless of which state the workflow is currently in.

    Args:
        submission_uuid (str): The submission associated with this workflow.

    Returns:
        None

    Raises:
        StaffAssessmentInternalError: Raised when there is an internal error
            creating the Workflow.

    """
    try:
        submission = submissions_api.get_submission_and_student(submission_uuid)
        StaffWorkflow.objects.get_or_create(
            course_id=submission['student_item']['course_id'],
            item_id=submission['student_item']['item_id'],
            submission_uuid=submission_uuid
        )
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while creating a new staff "
            "workflow for submission {}"
            .format(submission_uuid)
        )
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def on_cancel(submission_uuid):
    """
    Cancel the staff workflow for submission.

    Sets the cancelled_at field in staff workflow.

    Args:
        submission_uuid (str): The submission UUID associated with this workflow.

    Returns:
        None

    """
    try:
        workflow = StaffWorkflow.objects.get(submission_uuid=submission_uuid)
        workflow.cancelled_at = now()
        workflow.save(update_fields=['cancelled_at'])
    except StaffWorkflow.DoesNotExist:
        # If we can't find a workflow, then we don't have to do anything to
        # cancel it.
        pass
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while cancelling the staff"
            "workflow for submission {}"
            .format(submission_uuid)
        )
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def get_score(submission_uuid, staff_requirements):  # pylint: disable=unused-argument
    """
    Generate a score based on a completed assessment for the given submission.
    If no assessment has been completed for this submission, this will return
    None.

    Args:
        submission_uuid (str): The UUID for the submission to get a score for.
        staff_requirements (dict): Not used.

    Returns:
        A dictionary with the points earned, points possible,
        contributing_assessments, and staff_id information.

    """
    assessment = get_latest_staff_assessment(submission_uuid)
    if not assessment:
        return None

    return {
        "points_earned": assessment["points_earned"],
        "points_possible": assessment["points_possible"],
        "contributing_assessments": [assessment["id"]],
        "staff_id": assessment["scorer_id"],
    }


def get_latest_staff_assessment(submission_uuid):
    """
    Retrieve the latest staff assessment for a submission.

    Args:
        submission_uuid (str): The UUID of the submission being assessed.

    Returns:
        dict: The serialized assessment model
        or None if no assessments are available

    Raises:
        StaffAssessmentInternalError if there are problems connecting to the database.

    Example usage:

    >>> get_latest_staff_assessment('10df7db776686822e501b05f452dc1e4b9141fe5')
    {
        'points_earned': 6,
        'points_possible': 12,
        'scored_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 649284 tzinfo=<UTC>),
        'scorer': u"staff",
        'feedback': u''
    }

    """
    try:
        assessments = Assessment.objects.filter(
            submission_uuid=submission_uuid,
            score_type=STAFF_TYPE,
        )[:1]
    except DatabaseError as ex:
        msg = (
            "An error occurred while retrieving staff assessments "
            "for the submission with UUID {uuid}: {ex}"
        ).format(uuid=submission_uuid, ex=ex)
        logger.exception(msg)
        raise StaffAssessmentInternalError(msg) from ex

    if assessments:
        return full_assessment_dict(assessments[0])

    return None


def get_assessment_scores_by_criteria(submission_uuid):
    """Get the staff score for each rubric criterion

    Args:
        submission_uuid (str): The submission uuid is used to get the
            assessment used to score this submission.

    Returns:
        (dict): A dictionary of rubric criterion names, with a score of
            the staff assessments.

    Raises:
        StaffAssessmentInternalError: If any error occurs while retrieving
            information from the scores, an error is raised.
    """
    try:
        # This will always create a list of length 1
        assessments = list(
            Assessment.objects.filter(
                score_type=STAFF_TYPE, submission_uuid=submission_uuid
            )[:1]
        )
        scores = Assessment.scores_by_criterion(assessments)
        # Since this is only being sent one score, the median score will be the
        # same as the only score.
        return Assessment.get_median_score_dict(scores)
    except DatabaseError as ex:
        error_message = "Error getting staff assessment scores for {}".format(submission_uuid)
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def get_submission_to_assess(course_id, item_id, scorer_id):
    """
    Get a submission for staff evaluation.

    Retrieves a submission for assessment for the given staff member.

    Args:
        course_id (str): The course that we would like to fetch submissions from.
        item_id (str): The student_item (problem) that we would like to retrieve submissions for.
        scorer_id (str): The user id of the staff member scoring this submission

    Returns:
        dict: A student submission for assessment. This contains a 'student_item',
            'attempt_number', 'submitted_at', 'created_at', and 'answer' field to be
            used for assessment.

    Raises:
        StaffAssessmentInternalError: Raised when there is an internal error
            retrieving staff workflow information.

    Examples:
        >>> get_submission_to_assess("a_course_id", "an_item_id", "a_scorer_id")
        {
            'student_item': 2,
            'attempt_number': 1,
            'submitted_at': datetime.datetime(2014, 1, 29, 23, 14, 52, 649284, tzinfo=<UTC>),
            'created_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 668850, tzinfo=<UTC>),
            'answer': { ... }
        }

    """
    student_submission_uuid = StaffWorkflow.get_submission_for_review(course_id, item_id, scorer_id)
    if student_submission_uuid:
        try:
            submission_data = submissions_api.get_submission(student_submission_uuid)
            return submission_data
        except submissions_api.SubmissionNotFoundError as ex:
            error_message = (
                "Could not find a submission with the uuid {}"
            ).format(student_submission_uuid)
            logger.exception(error_message)
            raise StaffAssessmentInternalError(error_message) from ex
    else:
        logger.info(
            "No submission found for staff to assess ({}, {})"
            .format(
                course_id,
                item_id,
            )
        )
        return None


def get_staff_grading_statistics(course_id, item_id):
    """
    Returns the number of graded, ungraded, and in-progress submissions for staff grading.

    Args:
        course_id (str): The course that this problem belongs to
        item_id (str): The student_item (problem) that we want to know statistics about.

    Returns:
        dict: a dictionary that contains the following keys: 'graded', 'ungraded', and 'in-progress'
    """
    return StaffWorkflow.get_workflow_statistics(course_id, item_id)


def create_assessment(
        submission_uuid,
        scorer_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        scored_at=None
):
    # pylint: disable=unicode-format-string
    """
    Creates an assessment on the given submission.

    Assessments are created based on feedback associated with a particular
    rubric.

    Assumes that the user creating the assessment has the permissions to do so.

    Args:
        submission_uuid (str): The submission uuid for the submission being
            assessed.
        scorer_id (str): The user ID for the user giving this assessment. This
            is required to create an assessment on a submission.
        options_selected (dict): Dictionary mapping criterion names to the
            option names the user selected for that criterion.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        rubric_dict (dict): The rubric model associated with this assessment
        scored_at (datetime): Optional argument to override the time in which
            the assessment took place. If not specified, scored_at is set to
            now.

    Keyword Args:
        scored_at (datetime): Optional argument to override the time in which
            the assessment took place. If not specified, scored_at is set to
            now.

    Returns:
        dict: the Assessment model, serialized as a dict.

    Raises:
        StaffAssessmentRequestError: Raised when the submission_id is invalid, or
            the assessment_dict does not contain the required values to create
            an assessment.
        StaffAssessmentInternalError: Raised when there is an internal error
            while creating a new assessment.

    Examples:
        >>> options_selected = {"clarity": "Very clear", "precision": "Somewhat precise"}
        >>> criterion_feedback = {"clarity": "I thought this essay was very clear."}
        >>> feedback = "Your submission was thrilling."
        >>> create_assessment("Tim", options_selected, criterion_feedback, feedback, rubric_dict)
    """
    try:
        try:
            scorer_workflow = StaffWorkflow.objects.get(submission_uuid=submission_uuid)
        except StaffWorkflow.DoesNotExist:
            scorer_workflow = None

        assessment = _complete_assessment(
            submission_uuid,
            scorer_id,
            options_selected,
            criterion_feedback,
            overall_feedback,
            rubric_dict,
            scored_at,
            scorer_workflow
        )
        return full_assessment_dict(assessment)

    except InvalidRubric as ex:
        error_message = "The rubric definition is not valid."
        logger.exception(error_message)
        raise StaffAssessmentRequestError(error_message) from ex
    except InvalidRubricSelection as ex:
        error_message = "Invalid options were selected in the rubric."
        logger.warning(error_message, exc_info=True)
        raise StaffAssessmentRequestError(error_message) from ex
    except DatabaseError as ex:
        error_message = (
            "An error occurred while creating an assessment by the scorer with this ID: {}"
        ).format(scorer_id)
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


@transaction.atomic
def _complete_assessment(
        submission_uuid,
        scorer_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        scored_at,
        scorer_workflow
):
    """
    Internal function for atomic assessment creation. Creates a staff assessment
    in a single transaction.

    Args:
        submission_uuid (str): The submission uuid for the submission being
            assessed.
        scorer_id (str): The user ID for the user giving this assessment. This
            is required to create an assessment on a submission.
        options_selected (dict): Dictionary mapping criterion names to the
            option names the user selected for that criterion.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        rubric_dict (dict): The rubric model associated with this assessment
        scored_at (datetime): Optional argument to override the time in which
            the assessment took place. If not specified, scored_at is set to
            now.

    Returns:
        The Assessment model

    """
    # Get or create the rubric
    rubric = rubric_from_dict(rubric_dict)

    # Create the staff assessment
    assessment = Assessment.create(
        rubric,
        scorer_id,
        submission_uuid,
        STAFF_TYPE,
        scored_at=scored_at,
        feedback=overall_feedback
    )

    # Create assessment parts for each criterion in the rubric
    # This will raise an `InvalidRubricSelection` if the selected options do not
    # match the rubric.
    AssessmentPart.create_from_option_names(assessment, options_selected, feedback=criterion_feedback)

    # Close the active assessment
    if scorer_workflow is not None:
        scorer_workflow.close_active_assessment(assessment, scorer_id)
    return assessment
